import os
import ee
import time
from ee.ee_exception import EEException
import gcsfs
import logging
import datetime
from functools import partial
import warnings
import hydrafloods as hf
from hydrafloods import (
    datasets,
    timeseries,
    filtering,
    ml,
    utils,
    geeutils,
    indices,
    thresholding,
    decorators,
    corrections,
    fuzzy,
)


# set global constant variables in module that are used throughout
MERIT = ee.Image("MERIT/Hydro/v1_0_1")
DEM = MERIT.select("elv")
HAND = MERIT.select("hnd")
HAND_THRESH = 15  # hand threshold
HAND_MASK = HAND.lt(HAND_THRESH)
SLOPE = ee.Terrain.slope(HAND.unmask(0))
SAND = ee.Image("OpenLandMap/SOL/SOL_SAND-WFRACTION_USDA-3A1A1A_M/v02")


def hydro30(
    region,
    target_date,
    speckle_filter="gamma_map",
    grid_size=0.1,
    scale=90,
    combine_scenes=True,
):

    if not isinstance(target_date, ee.Date):
        target_date = ee.Date(target_date)

    s1 = datasets.Sentinel1(region, target_date, target_date.advance(1, "day"))

    filter_opts = dict(
        lee_sigma=filtering.lee_sigma,
        gamma_map=filtering.gamma_map,
        refined_lee=filtering.refined_lee,
    )
    filter_func = filter_opts[speckle_filter]

    proc = (
        (corrections.slope_correction, dict(elevation=DEM, buffer=60)),
        filter_func,
        (_water_classification, dict(grid_size=grid_size, scale=scale)),
    )

    processed = s1.pipe(proc)

    if combine_scenes:
        output = processed.collection.mosaic()
    else:
        output = processed.collection

    return output


@decorators.keep_attrs
def _water_classification(img, grid_size=0.1, scale=90):
    def _per_band_water(img, tiles, bandname):
        # !!! this function assumes one band image !!!
        def _calc_t(tile):
            histogram = band.updateMask(band.lt(0)).reduceRegion(
                ee.Reducer.histogram(max_buckets, min_bucket_width, max_raw)
                .combine("mean", None, True)
                .combine("variance", None, True),
                tile.geometry(),
                scale=scale,
                bestEffort=True,
                tileScale=16,
            )

            threshold = hf.otsu(histogram.get(bandname.cat("_histogram")))

            return tile.set("t", threshold)

        max_buckets = 255
        min_bucket_width = 0.001
        max_raw = 1e6
        bandname = ee.String(bandname)
        band = img.select(bandname)

        thresh_tiles = tiles.map(_calc_t)
        db_t = ee.Number(thresh_tiles.aggregate_array("t").reduce(ee.Reducer.mean()))

        water_init = band.lt(db_t)

        # db_stats = band.updateMask(band.lt(0)).reduceRegion(
        #     reducer=ee.Reducer.median(),
        #     geometry = region,
        #     scale=scale,
        #     bestEffort=True,
        #     tileScale=8
        # )
        #
        # db_lower = db_stats.get(bandname)
        # db_mid = db_t.add(db_lower).divide(2)

        fuzzy_db = fuzzy.fuzzy_large(band.select(bandname), db_t, 5)

        connected_objects = (
            water_init.select(bandname)
            .selfMask()
            .connectedComponents(connectedness=ee.Kernel.square(1.5), maxSize=1024)
        )

        connected_n = connected_objects.select("labels").connectedPixelCount(
            maxSize=1024
        )

        fuzzy_connected = fuzzy.fuzzy_zmf(connected_n.unmask(0), 10, 3)

        fuzziness = [
            fuzzy_slope,
            fuzzy_hand,
            fuzzy_db.select(bandname),
            fuzzy_connected,
            fuzzy_sand,
        ]

        weights = [1 / len(fuzziness) for _ in range(len(fuzziness))]

        fuzzy_combo = hf.fuzzy_weighted(fuzziness, weights).updateMask(
            band.select([0]).mask()
        )

        fuzzy_mask = fuzzy_combo.gt(0.60)

        water_final = water_init.And(fuzzy_mask)

        return water_final.rename(bandname.cat("_water")).uint8()

    def tile_stats(tile):
        stats = img.reduceRegion(
            reducer=stat_reducer,
            geometry=tile.geometry(),
            scale=scale,
            bestEffort=True,
            tileScale=16,
        )

        vv_denom = ee.Number(stats.get("VV_stdDev"))
        vh_denom = ee.Number(stats.get("VH_stdDev"))

        vv_denom = ee.Algorithms.If(vv_denom, vv_denom, ee.Number(0.00001))
        vh_denom = ee.Algorithms.If(vh_denom, vh_denom, ee.Number(0.00001))

        vv_ratio = ee.Number(img_stats.get("VV_p50")).divide(vv_denom)
        vh_ratio = ee.Number(img_stats.get("VH_p50")).divide(vh_denom)

        stats = stats.set("VV_ratio", vv_ratio).set("VH_ratio", vh_ratio)

        return tile.set(stats)

    #

    db_bands = img.bandNames().removeAll(["angle", "local_inc_angle"])
    hd_bands = ["hnd", "hand_mask"]
    img = img.select(db_bands).addBands(HAND_MASK).addBands(HAND)

    region = img.geometry()

    # this only creates tiles that are within the image so some tiles along edges will drop
    tiles = hf.tile_region(region, contain_geom=region, grid_size=grid_size)

    stat_reducer = ee.Reducer.stdDev().combine(ee.Reducer.mean(), None, True)

    img_stats = img.reduceRegion(
        reducer=ee.Reducer.percentile([50, 90]),
        geometry=region,
        scale=scale,
        bestEffort=True,
        tileScale=16,
    )

    stat_tiles = tiles.map(tile_stats).filter(ee.Filter.gte("hnd_mean", 0.8))

    vv_sel_tiles = stat_tiles.limit(5, "VV_ratio", False)
    vh_sel_tiles = stat_tiles.limit(5, "VH_ratio", False)

    # calculate the fuzzy membership of HAND Slope
    fuzzy_slope = fuzzy.fuzzy_zmf(SLOPE, 0, 15)

    hand_stats = HAND.updateMask(
        HAND.neq(0).And(HAND.lt(ee.Number(img_stats.get("hnd_p90"))))
    ).reduceRegion(
        reducer=ee.Reducer.median().combine(ee.Reducer.stdDev(), None, True),
        geometry=region,
        scale=scale,
        bestEffort=True,
        tileScale=16,
    )

    hand_lower = ee.Number(hand_stats.get("hnd_median"))
    hand_std = ee.Number(hand_stats.get("hnd_stdDev"))
    hand_upper = hand_lower.add(ee.Number(3.0).multiply(hand_std)).add(5.0)

    fuzzy_hand = fuzzy.fuzzy_zmf(HAND, hand_lower, hand_upper).unmask(1)

    fuzzy_sand = fuzzy.fuzzy_small(SAND.select([0]).unmask(0), 37.5, 8)

    vv_water = _per_band_water(img, vv_sel_tiles, "VV")
    vh_water = _per_band_water(img, vh_sel_tiles, "VH")

    overlap = vv_water.Or(vh_water)

    # post processing opening of the segmented water
    opened_closed_water = filtering.close_binary(filtering.open_binary(overlap))

    out_water = opened_closed_water.rename("water").updateMask(img.select([0]).mask())

    return out_water



import ee
from functools import partial

from hydrafloods import filtering


def fwdet(water_img, dem, band=None, outlier_test=False, force_projection=False):
    """Implementation of the Flood Water Depth Estimation Tool
    Used to calculate water depth given a water map and elevation data
    Original paper: https://doi.org/10.5194/nhess-19-2053-2019
    Earth Engine paper: https://doi.org/10.1109/LGRS.2020.3031190

    args:
        water_img (ee.Image): earth engine image object representing the water extent. Assumes values of 1 equal water.
        dem (ee.Image): earth engine image object representing elevation
        band (str | None): band name to use for algorithm, if set to `None` will use first band in image. default = None
        outlier_test (bool): flag used to find and filter outlier dem values around water boundaries. default = False
        force_projection (bool): flag used to force the output image projection to that of the input dem. default = False

    returns:
        ee.Image: image of water depth in meters
    """

    proj = dem.projection()
    res = proj.nominalScale().multiply(1.5)

    # select band to use for algorithm
    if band is None:
        water_img = water_img.select([0]).selfMask()

    else:
        water_img = water_img.select(ee.String(band)).selfMask()

    if outlier_test:
        filled = filtering.modified_median_zscore(dem)

        expand = water_img.focal_max(
            kernel=ee.Kernel.square(radius=res, units="meters")
        )

        dem_mask = dem.updateMask(water_img.gt(0))

        boundary = dem_mask.add(expand)

        filled = filtering.modified_median_zscore(boundary, filled)

    else:
        filled = dem

    # cumulative cost model

    mod = filled.updateMask(water_img.mask().eq(0))
    source = mod.mask()

    val = 10000
    push = 5000

    cost0 = ee.Image(val).where(source, 0).cumulativeCost(source, push)
    cost1 = ee.Image(val).where(source, 1).cumulativeCost(source, push)
    cost2 = mod.unmask(val).cumulativeCost(source, push)

    costFill = cost2.subtract(cost0).divide(cost1.subtract(cost0))

    costSurface = mod.unmask(0).add(costFill)

    # Kernel for low-pass filter
    boxcar = ee.Kernel.square(radius=3, units="pixels", normalize=True)

    # Floodwater depth calculation and smoothing using a low-pass filter
    depths = (
        costSurface.subtract(filled).convolve(boxcar).updateMask(water_img.mask())
    ).rename("depth")

    # force min values to be 0
    depths = depths.max(ee.Image.constant(0))

    # force the output project to be that of the input dem if force_projection is true
    if force_projection:
        depths = depths.reproject(proj)

    return depths


def fwdet_experimental(
    water_img,
    dem,
    iter=10,
    band=None,
    pixel_edge=0,
    boundary_definition="mean",
    smooth_depths=True,
    force_projection=False,
):
    """Experimantal changes to the implementation of the Flood Water Depth Estimation Tool
    Used to calculate water depth given a water map and elevation data

    Difference from original FwDET implementation include ...

    args:
        water_img (ee.Image): earth engine image object representing the water extent.
        dem (ee.Image): earth engine image object representing elevation
        iter (int|ee.Number, optional): keyword to set number of iterations to fill up every pixel.
            The algorithm will calculate minimum iterations needed and use the max between the two.
            default = 10
        band (str | None,optional): band name to use for algorithm, if set to `None` will use first band in image. default = None
        pixel_edge (int): value that represents the boundary/edge of water. Note: must be client-side int object. default = 0
        boundary_definition (str): method for defining the elevation at water edges. Options are 'mean' or 'nearest'
            Note: 'nearest' is the FwDET original method, 'mean' is an updated method. default=mean
        smooth_depths (bool): flag to control if the resulting water depths should be smoothed or not, uses a 3x3 pixel mean. default=True
        force_projection (bool): flag used to force the output image projection to that of the input dem. default = False


    returns:
        ee.Image: image of water depth in meters
    """
    proj = dem.projection()

    res = water_img.projection().nominalScale().multiply(1.5)

    # select band to use for algorithm
    if band is None:
        water_img = water_img.select([0])

    else:
        water_img = water_img.select(ee.String(band))

    # detect water edge
    watermap_edge = (
        water_img.selfMask().unmask(-999).focal_min(res, "square", "meters").eq(-999)
    )
    watermap_edge = watermap_edge.updateMask(water_img.unmask(0))
    watermap_edge = watermap_edge.selfMask()
    # watermap_edge = watermap_edge.reproject(water_img.projection());

    # detect watermap extent outer edge (not to be used in algorithm)
    if pixel_edge > 0:
        outer_edge = (
            water_img.mask()
            .unmask(0)
            .gte(0)
            .unmask(0, False)
            .gt(0)
            .focal_min(res.multiply(ee.Number(pixel_edge).add(1)), "square", "meters")
            .eq(0)
        )
        outer_edge = outer_edge.updateMask(outer_edge)
        # outer_edge = outer_edge.reproject(water_img.projection())
        # update watermap edge
        watermap_edge = watermap_edge.updateMask(outer_edge.unmask(0).eq(0))
        # watermap_edge = watermap_edge.reproject(water_img.projection());
    else:
        outer_edge = water_img.mask().unmask(0).lt(0)

    DEM = dem  # being super lazy here by setting a variable vs renaming things...

    # get elevation at edge
    # DEM_watered_edge = watermap_edge.multiply(DEM); # simply using boundary cell elevations
    DEM_masked = DEM.updateMask(
        water_img.unmask(0).eq(0).add(watermap_edge.unmask(0)).eq(1)
    )
    DEM_watered_edge = watermap_edge.multiply(
        DEM_masked.focal_mean(res, "square", "meters")
    )
    DEM_watered_edge = DEM_watered_edge  # .reproject(water_img.projection())
    # calculate (approximation of) minimum needed iterations to fill up every pixel
    req_iter = (
        water_img.eq(0)
        .add(watermap_edge)
        .fastDistanceTransform(50, "pixels", "squared_euclidean")
        .sqrt()
        .round()
        .updateMask(water_img)
        .updateMask(outer_edge.unmask(0).eq(0))
        .reproject(water_img.projection())
    )
    req_iter = req_iter.reduceRegion(
        reducer=ee.Reducer.max(),
        geometry=water_img.geometry().bounds(1e3),
        scale=res,
        maxPixels=1e12,
    ).get("distance")
    # set iterations to be used
    iter = ee.Number(req_iter).max(iter)
    # construct list to iterate over (list contents unused, just for iterations)
    iter_list = ee.List.sequence(0, iter)

    # use look up which algorithm to use for defining the boundary elevation values
    boundary_algos = {"mean": _mbce, "nearest": _nbce}
    f_boundary = partial(
        boundary_algos[boundary_definition], mask=water_img, resolution=res
    )

    # apply boundary condition algo, starting with edge values and filling it up further with each iteration
    DEM_watered_fill = ee.Image(iter_list.iterate(f_boundary, DEM_watered_edge))
    DEM_watered = DEM_watered_fill.reproject(water_img.projection())
    # derive water depths
    depths = DEM_watered.subtract(DEM).max(0)

    # smooth water depths
    # Cohen et al. (2018) use a 3x3 pixel mean (but FwDET-GEE seems to use a boxcar kernel?)
    if smooth_depths:
        depths = depths.reduceNeighborhood(
            reducer=ee.Reducer.mean(),
            kernel=ee.Kernel.square(res, units="meters"),
            optimization="boxcar",  # optimization for mean.
        )

    # force the output project to be that of the input dem if force_projection is true
    if force_projection:
        depths = depths.reproject(proj)

    return depths.rename("depth").set("depth_iter", iter)


# Mean Boundary Cell Elevation (MBCE), adapted from NBCE above
# MBCE function
def _mbce(i, img, mask=None, resolution=30):
    """Mean Boundary Cell Elevation (MBCE), unpublished A. Haag (2021)"""
    # obtain mean boundary cell elevation
    new_img = ee.Image(img).focal_mean(resolution, "square", "meters")
    # make sure the original / previous step values are kept intact (comment out for more smoothing)
    # new_img = new_img.where(ee.Image(img).gt(-999), ee.Image(img));
    # make sure it does not expand to other side of floodmap (where there is no water)
    return new_img.updateMask(mask.unmask(0))  # .reproject(floodmap.projection())


# Nearest Boundary Cell Elevation (NBCE) algorithm
def _nbce(i, img, mask=None, resolution=30):
    """Nearest Boundary Cell Elevation (NBCE), Cohen et al. (2018)"""
    # obtain nearest boundary cell elevations using different kernels
    new_img_0 = ee.Image(img).focal_min(resolution, "square", "meters")
    new_img_1 = ee.Image(img).focal_min(resolution, "plus", "meters")
    new_img_2 = ee.Image(img).focal_min(resolution, "cross", "meters")
    # make sure the original / previous step values are kept intact
    new_img_0 = new_img_0.where(ee.Image(img).gt(-999), ee.Image(img))
    new_img_1 = new_img_1.where(ee.Image(img).gt(-999), ee.Image(img))
    new_img_2 = new_img_2.where(ee.Image(img).gt(-999), ee.Image(img))
    # add the NBCE in right order (minimum distance at the end, overwrites diagonal)
    new_img = new_img_0.where(new_img_2.gt(-999), new_img_2)
    new_img = new_img.where(new_img_1.gt(-999), new_img_1)
    # make sure it does not expand to other side of floodmap (where there is no water)
    return new_img.updateMask(mask.unmask(0))


def downscale_wfraction(fraction, dem):
    """Algorithm to downscale water fraction maps using DEM data
    Converts fractional water data to higher resolution binary water data
    Paper: https://doi.org/10.1016/j.rse.2013.03.015

    args:
        fraction (ee.Image): earth engine image object representing the fractional water extent within each pixel.
        dem (ee.Image): earth engine image object representing elevation

    returns:
        ee.Image: downscaled binary image of water
    """

    resolution = fraction.projection().nominalScale()

    water_present = fraction.gt(0.0)

    h_min = dem.updateMask(water_present).focal_min(resolution, "square", "meters")
    h_max = dem.updateMask(water_present).focal_max(resolution, "square", "meters")

    water_high = h_min.add(h_max.subtract(h_min).multiply(fraction))

    return dem.gte(h_min).And(dem.lt(water_high))



s1 = {
    "edge_otsu": {
        "initial_threshold": -16,
        "edge_buffer": 300,
        "scale": 150,
        "band": "VV",
    },
    "bmax_otsu": {
        "initial_threshold": -16,
        "grid_size": 0.25,
        "scale": 150,
        "band": "VV",
    },
}

lc8 = {
    "edge_otsu": {
        "initial_threshold": 0.0,
        "edge_buffer": 300,
        "scale": 150,
        "invert": True,
        "band": "mndwi",
    },
    "bmax_otsu": {"initial_threshold": 0.0, "scale": 150, "invert": True},
}

fusion = {
    "edge_otsu": {
        "initial_threshold": 0.0,
        "edge_buffer": 300,
        "scale": 150,
        "invert": True,
    },
    "bmax_otsu": {"initial_threshold": 0.0, "scale": 150, "invert": True},
}



from __future__ import print_function, division
import datetime
import ee
import os
import sys
import subprocess
import gcsfs


def list_gcs_objs(bucket_path, pattern=None, output_url=False, project=None):
    """Function to list objects in Google Cloud Storage Bucket

    args:
        bucket_path (str): Google Cloud Storage bucket name
        pattern (str | None, optional): regex pattern to search in bucket. 
            Can seach folders by adding folder names (i.e. pattern = 'subfolder/*.txt).
            If None then will not use search pattern. default = None
        output_url (bool, optional): boolean switch to output google cloud storage http url
            or google cloud storage object uri. If false will output gcs uri. default = False
        project (str | None): Cloud project name to use when initiation file spec. If None then
            use default gcloud config. default = None

    returns:
        list[str]: List of objects in bucket that match pattern
    """
    fs = gcsfs.GCSFileSystem(project=project)
    if pattern is not None:
        bucket_path = (
            bucket_path + "/" if not bucket_path.endswith("/") else bucket_path
        )
        blobs = fs.glob(f"{bucket_path}{pattern}")
    else:
        blobs = fs.ls(bucket_path)

    base = "https://storage.cloud.google.com/{0}" if output_url else "gs://{0}"

    return [base.format(blob) for blob in blobs]


def push_to_gcs(file, bucket_path):
    """Helper function to copy local files to Google Cloud Storage
    Thinly wraps `gsutil cp` command line

    args:
        file (str): file path to push to GCS
        bucket_path (str): path on GCS to copy file to

    """
    if os.path.exists(file):
        cmd = "gsutil cp {0} {1}".format(file, bucket_path)
        proc = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        out, err = proc.communicate()
    else:
        raise ValueError('file "{0} does not exist'.format(file))
    return


def push_to_ee(bucket_obj, asset_collection, properties=None, delete_bucket_obj=False):
    """Helper function to begin ingest process for imagery on GCS to GEE
    Thinly wraps `earthengine upload image`

    args:
        bucket_obj (str): GCS bucket object to ingest into GEE. Expects that object has mime type of image/tiff
        asset_collection (str): Earth Engine asset collection to push object to
        properties (list[str], optional): list of properties to set when ingesting files. If None then no properties
            will be set. default = None
        delete_bucket_obj (bool, optional): boolean switch to delete GCS object once ingested into EE. If set to False
            then file will remain on GCS. default = False
    """
    name = os.path.basename(bucket_obj).replace(".", "_")
    asset = asset_collection + name

    pStr = ""
    for i in properties:
        pStr += "--{0} {1} ".format(i, properties[i])

    binPath = os.path.dirname(sys.executable)
    cmd = "{0}/earthengine upload image --asset_id={1} {2} {3}".format(
        binPath, asset, pStr, bucket_obj
    )
    proc = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    out, err = proc.communicate()
    if properties:
        pStr = ""

        running = True
        while running == True:
            tasks = ee.batch.Task.list()
            if "COMPLETED" in str(tasks[0]):
                running = False
            elif "FAILED" in str(tasks[0]):
                print(
                    "EE upload process failed for image {}, check Earth Engine for error".format(
                        bucket_obj
                    )
                )
                sys.exit(1)

    if delete_bucket_obj:
        cmd = "gsutil rm {0}".format(bucket_obj)
        proc = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        out, err = proc.communicate()

    return


def decode_date(date):
    """Decodes a date from a command line argument, returning msec since epoch".
    
    args:
        date (str): date value in a format that can be parsed into datetime object
        
    returns:
        datetime.datetime: decoded datetime value

    raises:
        TypeError: if string does not conform to a legal date format.
    """

    date_formats = [
        "%Y%m%d",
        "%Y-%m-%d",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%f",
    ]
    for date_format in date_formats:
        try:
            dt = datetime.datetime.strptime(date, date_format)
            return dt
        except ValueError:
            continue
    raise TypeError(f"Invalid value for property of type 'date': '{date}'.")




from __future__ import print_function, division
import ee
from hydrafloods import decorators, ml, thresholding


def starfm(
    coarseCollection, fineCollection=None, targetDate="1970-01-01", windowSize=33, A=0.5
):
    @decorators.keep_attrs
    def apply_starfm(img):
        t = ee.Date(img.get("system:time_start"))

        Tijk = img.select("time").subtract(base.select("time_first"))

        Sijk = img.subtract(slv).abs().convolve(Dijk).rename(bandList)

        Cijk = Sijk.multiply(Tijk).convolve(Dijk)
        Wijk = one.divide(Cijk).divide(one.divide(Cijk).convolve(Dijk))

        nBands = ee.Number(base.bandNames().length())
        expected = ee.Number(len(bandList))

        outImg = ee.Algorithms.If(
            nBands.neq(expected),
            None,
            base.divide(Wijk.convolve(Dijk))
            .add(Sijk)
            .set("system:time_start", t)
            .rename(bandList),
        )

        return ee.Image(outImg)

    target = ee.Date(targetDate)

    if type(fineCollection) is not ee.imagecollection.ImageCollection:
        try:
            fineCollection = getattr(fineCollection, "collection")
        except Exception as e:
            raise TypeError(
                "keyword fineCollection needs to be either of type ee.ImageCollection "
                "or hf.hfCollection"
            )

    bandList = ee.Image(coarseCollection.first()).bandNames().getInfo()

    one = ee.Image.constant(1)
    centerPos = ee.Number((windowSize - 1) / 2)
    pos = ee.Array(ee.List.sequence(0, windowSize - 1))
    xPos = pos.repeat(1, windowSize)
    yPos = pos.repeat(1, windowSize).transpose()

    dijk = (
        ee.Array(centerPos)
        .subtract(xPos)
        .pow(2)
        .add(ee.Array(centerPos).subtract(yPos).pow(2))
        .sqrt()
    )

    weightMax = dijk.get([0, 0]).add(1)
    # flip weights to have heighest in middle
    # normalize values for convolutions
    dijk_norm = dijk.subtract(weightMax).abs().divide(weightMax.subtract(1))

    dW = ee.Array(1).add(dijk.divide(ee.Array(A)))

    square = ee.Kernel.square(windowSize, "pixels")
    Dijk = ee.Kernel.fixed(
        windowSize, windowSize, dW.toList(), centerPos, centerPos, True
    )

    base = (
        fineCollection.filterDate(target.advance(-2, "month"), target)
        .sort("system:time_start", False)
        .reduce(ee.Reducer.firstNonNull())
    )

    slv = (
        coarseCollection.filterDate(target.advance(-2, "month"), target)
        .sort("system:time_start", False)
        .reduce(ee.Reducer.firstNonNull())
    )

    result = coarseCollection.filterDate(
        target.advance(-5, "day"), target.advance(5, "day")
    ).map(apply_starfm, True)

    return result


def bathtub(wfrac, hand, permanent=None):
    """
    Function to fit hi-resolution HAND model to water fraction estimate
    args:
        wFrac (ee.Image): water fraction image, values must be 0-1
        hand (ee.Image): height above nearest drainage (HAND) image, units in meters
        permanent (ee.Image): permanent water image to seed HAND filling
    """

    def fillGrids(d):
        d = ee.Number(d)
        dMap = (
            hand.lte(d)
            .Or(permWater)
            .reduceResolution(
                reducer=ee.Reducer.mean(), bestEffort=True, maxPixels=1024
            )
            .reproject(crs=proj)
        )

        diff = wfrac.subtract(dMap).abs()
        return diff

    def minimizeDepth(d):
        d = ee.Number(d)
        dMap = hand.lte(d).Or(permWater)
        rd = dMap.reduceResolution(
            reducer=ee.Reducer.mean(), bestEffort=True, maxPixels=1024
        ).reproject(crs=proj)

        diff = wfrac.subtract(rd).abs()

        out = nil.where(diff.eq(minDiff), dMap)
        err = nil.where(diff.eq(minDiff), diff)
        return out.rename("water").addBands(err.rename("error"))

    nil = ee.Image(0)
    if permanent:
        permWater = permanent
    else:
        permWater = ee.Image(0)
    proj = wfrac.projection()

    depths = ee.List.sequence(0, 20)

    depthMaps = ee.ImageCollection(depths.map(fillGrids))
    minDiff = depthMaps.min()

    waterMap = ee.ImageCollection(depths.map(minimizeDepth))
    final = (
        waterMap.select("water").max().addBands(waterMap.select("error").min().float())
    )

    return final


@decorators.keep_attrs
def pca_fusion(
    img,
    img_stats=None,
    match_stats=None,
    eigen_vecs=None,
    scaling_dict=None,
    img_band=None,
    match_band=None,
    inverse_relationship=True,
    reverse_pca_bands=False,
):
    pca_bands = [img_band, match_band]

    if reverse_pca_bands:
        pca_bands = pca_bands[::-1]

    img_z = (
        img.select(img_band)
        .subtract(img_stats.select(img_band + "_mean"))
        .divide(img_stats.select(img_band + "_stdDev"))
    )

    if inverse_relationship:
        img_z = img_z.multiply(-1)

    match_synth = (
        match_stats.select(match_band + "_mean")
        .add(img_z.multiply((match_stats.select(match_band + "_stdDev"))))
        .rename(match_band)
    )

    if reverse_pca_bands:
        input_img = match_synth.addBands(img)
    else:
        input_img = img.addBands(match_synth)

    if scaling_dict is not None:
        input_img = ml.standard_image_scaling(input_img, scaling_dict, pca_bands)

    pcs = ml.apply_image_pca(input_img, eigen_vecs, pca_bands)

    df_index = pcs.select([0]).clamp(-3.75, 3.75).rename("df_index")

    return df_index

@decorators.keep_attrs
def dnns(img, region=None):
    """Dynamic Nearest Neighbor Search algorithm for estimating 
    
    """
    pimg = img.select(["red","nir","swir1"])

    kernel_size = 40  #pixels?
    scale = pimg.projection().nominalScale()

    kernel = ee.Kernel.square(kernel_size, "pixels", False)
    kernel_norm = ee.Kernel.square(kernel_size, 'pixels', True)

    water = img.select("mndwi").gt(0.15)#hf.bmax_otsu(img, band="mndwi", initial_threshold=0.1, region=region).Not()
    land = img.select("mndwi").lt(-0.15)

    mix = water.Not().And(land.Not())
    ave_water = water.multiply(pimg).reduceRegion(ee.Reducer.mean(),region,scale,maxPixels=1e4,bestEffort=True).toImage()

    N_nmin_water = 1
    N_water = water.convolve(kernel)

    water_rf = (
        water.multiply(pimg)
        .convolve(kernel)
        .multiply(N_water.gte(N_nmin_water))
        .divide(N_water)
    )
    water_rf = water_rf.add(ave_water.multiply(water_rf.Not()))

    ave_land = pimg.multiply(land).reduceRegion(ee.Reducer.mean(),region,scale,maxPixels=1e4,bestEffort=True).toImage()

    R1 = pimg.select("red").divide(pimg.select("swir1"))
    R2 = pimg.select("nir").divide(pimg.select("swir1"))
    R3 = R1.subtract(water_rf.select("red").divide(pimg.select("swir1")))
    R4 = R2.subtract(water_rf.select("red").divide(pimg.select("swir1")))

    NR1 = R1.neighborhoodToBands(kernel)
    NR2 = R2.neighborhoodToBands(kernel)
    NI1 = img.select("red").neighborhoodToBands(kernel)
    NI2 = img.select("nir").neighborhoodToBands(kernel)
    NI3 = img.select("swir1").neighborhoodToBands(kernel)

    M1 = (NR1.gt(R3)).And(NR1.lt(R1))
    M2 = (NR2.gt(R4)).And(NR2.lt(R2))
    nLP = M1.And(M2)

    NnLP = nLP.reduce(ee.Reducer.sum())
    ave_nI1 = NI1.multiply(nLP).reduce(ee.Reducer.sum()).divide(NnLP)
    ave_nI2 = NI2.multiply(nLP).reduce(ee.Reducer.sum()).divide(NnLP)
    ave_nI3 = NI3.multiply(nLP).reduce(ee.Reducer.sum()).divide(NnLP)

    N_nmin_land = 1
    ave_pureland = ee.Image.cat([ave_nI1,ave_nI2,ave_nI3])
    ave_pureland = ave_pureland.multiply(NnLP.gte(N_nmin_land)).add(
        ave_land.multiply(NnLP.lt(N_nmin_land))
    )

    ave_landI3 = ave_land.select([2])
    f_water_all = (
        (ave_pureland.subtract(pimg))
        .divide(ave_pureland.subtract(water_rf))
        .clamp(0, 1)
    )
    f_water = f_water_all.add(water).subtract(land).clamp(0, 1)

    return f_water.reduce(ee.Reducer.mean()).rename("water_fraction").toFloat()



import ee

from hydrafloods import (
    timeseries,
    decorators,
    geeutils,
    thresholding
)


@decorators.keep_attrs
def extract_flood(observation, reference="seasonal", permanent_threshold=75):
    """Function used to extract flooded area based off of different JRC datasets.
    Expects that the water image has water = 1, land = 0, unknown = mask

    args:
        observation (ee.Image): image object representing the observation of water to extract floods
        reference (str): string to define how permanent water is defined. 'yearly' means use past 5 years
            of JRC yearly data and extract permanent class. 'seasonal' means use monthly water occurrence
            for the month of observation. 'occurrence' means use full JRC occurrence dataset irregardless of
            time. default = seasonal
        permanent_threshold (int): threshold value in % to define permanent water. Only used when
            reference equals 'seasonal' or 'occurrence'. default = 75

    returns:
        ee.Image: the extracted flood image where floods = 1

    raises:
        NotImplementedError: when user provides an invalid option for reference parameter
    """

    if reference == "yearly":
        end_time = observation.date()
        permanent_water = (
            ee.ImageCollection(
                "JRC/GSW1_2/YearlyHistory"
            )  # get the JRC historical dataset
            .filterDate(
                "1985-01-01", end_time
            )  # filter for historical data up to date of interest
            .limit(5, "system:time_start", False)  # grab the 5 latest images
            .map(
                lambda x: x.select("waterClass").eq(3)
            )  # extract out the permanent water class
            .sum()  # check if a pixel has been classified as permanent water in the past 5 years
            .unmask(0)
            .gt(0)
        )

    elif reference == "seasonal":
        month = observation.date().get("month")
        permanent_water = (
            ee.ImageCollection("JRC/GSW1_3/MonthlyRecurrence")
            .select("monthly_recurrence")
            .filter(ee.Filter.eq("month", month))
            .first()
            .gt(permanent_threshold)
        )

    elif reference == "occurrence":
        permanent_water = (
            ee.Image("JRC/GSW1_3/GlobalSurfaceWater")
            .select("occurrence")
            .gt(permanent_threshold)
        )

    else:
        raise NotImplementedError(
            "the selected reference method, {reference}, is not implemeted. please select either 'yearly','seasonal', or 'occurrence'"
        )

    return discrete_difference(observation, permanent_water)


@decorators.keep_attrs
def discrete_difference(observation, reference):
    """Function to find the difference between an observation and reference dataset.
    The data should be discrete values (i.e. a classification). This function
    expects that water/floods = 1 and land = 0. Unknown values should be masked.

    args:
        observation (ee.Image): image object representing the observation of water to extract floods
        reference (ee.Image): image object representing reference/permanent water to extract floods from

    returns:
        ee.Image: extracted flood image where flood values = 1
    """
    og_mask = observation.mask()
    floods = observation.unmask(0).add(reference.unmask(0).multiply(2)).eq(1)
    return floods.updateMask(og_mask).rename("flood")


@decorators.keep_attrs
def lar_change_detection(
    observation, reference, band=None, in_units="dB", segmentation="edgeotsu", **kwargs
):
    """Log Amplitude Ratio change detection method. https://doi.org/10.1080/014311698215649
    Note: This method only works for SAR imagery.

    args:
        observation (ee.Image):
        reference (ee.Image):
        band (str | None,optional): band name to use for thresholding, if set to `None` will use first band in image. default = None
        in_units (str, optional): string specifying the input units for the imagery. Options are 'dB' or 'power'. If in_units = 'dB',
            then the data is converted to power units. default = dB
        segmentation (str | None, optional): segmentation method to use to . Options are 'edgeotsu', 'bmaxotsu', or None. If none
            is provided then no segmentation is applied and the raw log amplitude ratio is returned. default = edgeotsu
        **kwargs: optional keywords to pass to segmentation method, not required if segmentation = None


    returns:
        ee.Image:
    """
    if band is None:
        band = observation.bandNames().get(0)
    # convert the db data to amplitude units
    # amplitude = sqrt(power)
    # then divide post/pre and take the log
    if in_units == "dB":
        lar = (
            geeutils.db_to_power(observation).sqrt().select(band)
            .divide(geeutils.db_to_power(reference).sqrt().select(band))
            .log10()
        )

    elif in_units == "power":
        lar = observation.sqrt().select(band).divide(reference.sqrt().select(band)).log10()

    else:
        raise NotImplementedError(
            f"input units could not be infered from {in_units}, please select either 'dB' or 'power'"
        )

    if segmentation == "edgeotsu":
        floods_lar = thresholding.edge_otsu(lar, band=band, **kwargs)

    elif segmentation == "bmaxotsu":
        floods_lar = thresholding.bmax_otsu(lar, band=band, **kwargs)

    elif segmentation == None:
        floods_lar = lar

    else:
        raise NotImplementedError(
            f"selected segmentation method {segmentation} is not valid, please select 'edgeotsu', 'bmaxotsu', or None"
        )

    return floods_lar


# def temporal_zscore_change_detection(
#     collection,
#     reference_period,
# ):
#     """Flood dectection method from https://doi.org/10.1016/j.rse.2020.111664
#     Calculates the per-pixel z_score from an dryseason composite and determines
#     flooded pixels using a decision tree
#     """
#
#     return


def flood_duration(collection, is_masked=True):

    if not isinstance(collection, ee.ImageCollection):
        collection = collection.collection

    max_extent = collection.max().selfMask()

    if is_masked == True:
        f = lambda x: timeseries.add_time_band(
            x.selfMask(), apply_mask=True, offset="day"
        )

    else:
        f = lambda x: timeseries.add_time_band(x, apply_mask=True, offset="day")

    time_coll = collection.map(f)

    min_time = time_coll.select("time").min()
    max_time = time_coll.select("time").max()

    return max_time.subtract(min_time).updateMask(max_extent)



import ee
from ee.ee_exception import EEException
import random
from hydrafloods import geeutils, decorators, ml, fuzzy


@decorators.keep_attrs
def bmax_otsu(
    img,
    band=None,
    region=None,
    scale=90,
    initial_threshold=0,
    thresh_no_data=None,
    invert=False,
    grid_size=0.1,
    bmax_threshold=0.75,
    iters=1,
    max_boxes=100,
    seed=7,
    max_buckets=255,
    min_bucket_width=0.001,
    max_raw=1e6,
    return_threshold=False,
):
    """Implementation of the B-Max Otsu thresholding algorithm.
    Detailed explanation of algorithm can be found at https://doi.org/10.3390/rs12152469

    args:
        img (ee.Image): input image to thresholding algorithm
        band (str | None,optional): band name to use for thresholding, if set to `None` will use first band in image. default = None
        region (ee.Geometry | None, optional): region to determine threshold, if set to `None` will use img.geometry(). default = None
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90
        initial_threshold (float, optional): initial estimate of water/no-water for estimating the probabilities of classes in segment. default = 0
        thresh_no_data (float, optional): threshold to be used when histogram is empty (likely little to no water present in image). default = -0.2
        invert (bool, optional): boolean switch to determine if to threshold greater than (True) or less than (False). default = False
        grid_size (float, optional): size in decimal degrees to tile image/region to check for bimodality. default = 0.1
        bmax_threshold (float, optional): value 0-1 to determine if a value of bmax is bimodal or not. default = 0.75
        iters (int, optional): number of iterations to successively shrink the bmax tiles to refine threshold. 1 uses defined grid_size. default = 1
        max_boxes (int, optional): maximum number of tiles/boxes to use when determining threshold, will select based on maximum bmax ordering. default = None
        seed (int, optional): random number generator seed for randomly selected max_boxes. default = 7
        max_buckets (int, optional): The maximum number of buckets to use when building a histogram; will be rounded up to a power of 2. default = 255
        min_bucket_width (float, optional): The minimum histogram bucket width to allow any power of 2. default = 0.001
        max_raw (int, optional): The number of values to accumulate before building the initial histogram. default = 1e6
        return_threshold (bool, optional): boolean switch, if set to true then function will return threshold number, else thresholded image. default = False

    returns:
        ee.Image: thresholded image based (if return_threshold==False) or threshold value (if return_threshold==True) based on the threshold determined by the algorithm
    """

    def calcBmax(feature):
        """Closure function to calculate Bmax for each feature covering image"""
        segment = img
        initial = segment.lt(initial_threshold)
        p1 = ee.Number(
            initial.reduceRegion(
                reducer=ee.Reducer.mean(),
                geometry=feature.geometry(),
                bestEffort=True,
                scale=scale,
            ).get(histBand)
        )
        p1 = ee.Number(ee.Algorithms.If(p1, p1, 0.9999))
        p2 = ee.Number(1).subtract(p1)

        m = (
            segment.updateMask(initial)
            .rename("m1")
            .addBands(segment.updateMask(initial.Not()).rename("m2"))
        )

        mReduced = m.reduceRegion(
            reducer=ee.Reducer.mean(),
            geometry=feature.geometry(),
            bestEffort=True,
            scale=scale,
        )

        m1 = ee.Number(mReduced.get("m1"))
        m2 = ee.Number(mReduced.get("m2"))

        m1 = ee.Number(ee.Algorithms.If(m1, m1, -25))
        m2 = ee.Number(ee.Algorithms.If(m2, m2, 0))

        sigmab = p1.multiply(p2.multiply(m1.subtract(m2).pow(2)))
        sigmat = ee.Number(
            segment.reduceRegion(
                reducer=ee.Reducer.variance(),
                geometry=feature.geometry(),
                bestEffort=True,
                scale=scale,
            ).get(histBand)
        )
        sigmat = ee.Number(ee.Algorithms.If(sigmat, sigmat, 2))
        bmax = sigmab.divide(sigmat)
        return feature.set({"bmax": bmax})

    if band is None:
        img = img.select([0])
        histBand = ee.String(img.bandNames().get(0))

    else:
        histBand = ee.String(band)
        img = img.select(histBand)

    if region is None:
        region = img.geometry()

    grid = geeutils.tile_region(region, centroid_within=region, grid_size=grid_size)

    bmaxes = grid.map(calcBmax).filter(ee.Filter.gt("bmax", bmax_threshold))

    if iters > 1:
        for i in range(iters - 1):
            grid_size /= 2.0
            grid = geeutils.tile_region(
                bmaxes.geometry(), centroid_within=bmaxes, grid_size=grid_size
            )

            bmaxes = grid.map(calcBmax).filter(ee.Filter.gt("bmax", bmax_threshold))

    if max_boxes is not None:
        selection = bmaxes.limit(max_boxes, "bmax")
    else:
        selection = bmaxes

    histogram = img.reduceRegion(
        ee.Reducer.histogram(max_buckets, min_bucket_width, max_raw)
        .combine("mean", None, True)
        .combine("variance", None, True),
        selection,
        scale,
        bestEffort=True,
        tileScale=16,
    )

    if thresh_no_data is not None:
        threshold = ee.Number(
            ee.Algorithms.If(
                ee.Dictionary(histogram.get(histBand.cat("_histogram"))).contains(
                    "bucketMeans"
                ),
                otsu(histogram.get(histBand.cat("_histogram"))),
                thresh_no_data,
            )
        )
    else:
        threshold = otsu(histogram.get(histBand.cat("_histogram")))

    if return_threshold is True:
        return ee.Image(threshold)
    else:
        water = ee.Image(ee.Algorithms.If(invert, img.gt(threshold), img.lt(threshold)))
        return water.rename("water").uint8()


@decorators.keep_attrs
def edge_otsu(
    img,
    band=None,
    region=None,
    scale=90,
    initial_threshold=0,
    thresh_no_data=None,
    invert=False,
    canny_threshold=0.05,
    canny_sigma=0,
    canny_lt=0.05,
    connected_pixels=200,
    edge_length=50,
    edge_buffer=100,
    max_buckets=255,
    min_bucket_width=0.001,
    max_raw=1e6,
    return_threshold=False,
):
    """Implementation of the Edge Otsu thresholding algorithm.
    Detailed explanation of algorithm can be found at https://doi.org/10.3390/rs12152469

    args:
        img (ee.Image): input image to thresholding algorithm
        band (str | None,optional): band name to use for thresholding, if set to `None` will use first band in image. default = None
        region (ee.Geometry | None, optional): region to determine threshold, if set to `None` will use img.geometry(). default = None
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90
        initial_threshold (float, optional): initial estimate of water/no-water for estimating the edges. default = 0
        thresh_no_data (float, optional): threshold to be used when histogram is empty (likely little to no water present in image). default = -0.2
        invert (bool, optional): boolean switch to determine if to threshold greater than (True) or less than (False). default = False
        canny_threshold (float, optional): threshold for canny edge detection. default = 0.05
        canny_sigma (float, optional): sigma value for gaussian filter in canny edge detection. default = 0
        canny_lt (float, optional): lower threshold for canny detection. default = 0.05
        connected_pixels (int, optional): maximum size of the neighborhood in pixels to determine if connected. default = 200
        edge_length (int, optional): minimum length of edges from canny detection to be considered edge. default = 50
        edge_buffer (int, optional): distance in meters to buffer edges on a side for histogram sampling. default = 100
        max_buckets (int, optional): The maximum number of buckets to use when building a histogram; will be rounded up to a power of 2. default = 255
        min_bucket_width (float, optional): The minimum histogram bucket width to allow any power of 2. default = 0.001
        max_raw (int, optional): The number of values to accumulate before building the initial histogram. default = 1e6
        return_threshold (bool, optional): boolean switch, if set to true then function will return threshold number, else thresholded image. default = False

    returns:
        ee.Image: thresholded image based (if return_threshold==False) or threshold value (if return_threshold==True) based on the threshold determined by the algorithm
    """

    if band is None:
        img = img.select([0])
        histBand = ee.String(img.bandNames().get(0))

    else:
        histBand = ee.String(band)
        img = img.select(histBand)

    if region is None:
        region = img.geometry()

    binary = img.lt(initial_threshold).rename("binary")

    # get canny edges
    canny = ee.Algorithms.CannyEdgeDetector(binary, canny_threshold, canny_sigma)
    # process canny edges
    connected = (
        canny.mask(canny).lt(canny_lt).connectedPixelCount(connected_pixels, True)
    )
    edges = connected.gte(edge_length)
    edgeBuffer = edges.focal_max(edge_buffer, "square", "meters")

    # mask out areas to get histogram for Otsu
    histogram_image = img.updateMask(edgeBuffer)

    histogram = histogram_image.reduceRegion(
        ee.Reducer.histogram(max_buckets, min_bucket_width, max_raw)
        .combine("mean", None, True)
        .combine("variance", None, True),
        region,
        scale,
        bestEffort=True,
        tileScale=16,
    )

    if thresh_no_data is not None:
        threshold = ee.Number(
            ee.Algorithms.If(
                ee.Dictionary(histogram.get(histBand.cat("_histogram"))).contains(
                    "bucketMeans"
                ),
                otsu(histogram.get(histBand.cat("_histogram"))),
                thresh_no_data,
            )
        )
    else:
        threshold = otsu(histogram.get(histBand.cat("_histogram")))

    if return_threshold is True:
        return threshold
    else:
        water = ee.Image(ee.Algorithms.If(invert, img.gt(threshold), img.lt(threshold)))
        return water.rename("water").uint8()


def otsu(histogram):
    """Otsu's method threhsolding algorithm.
    Computes single intensity threshold that separate histogram into two classes, foreground and background

    args:
        histogram (ee.Dictionary): computed object from ee.Reducer.histogram with keys "histogram" and "bucketMeans"

    returns:
        ee.Number: value of maximum inter-class intensity variance based on histogram
    """
    counts = ee.Array(ee.Dictionary(histogram).get("histogram"))
    means = ee.Array(ee.Dictionary(histogram).get("bucketMeans"))
    size = means.length().get([0])
    total = counts.reduce(ee.Reducer.sum(), [0]).get([0])
    sums = means.multiply(counts).reduce(ee.Reducer.sum(), [0]).get([0])
    mean = sums.divide(total)
    indices = ee.List.sequence(1, size)
    # Compute between sum of squares, where each mean partitions the data.

    def bss_function(i):
        aCounts = counts.slice(0, 0, i)
        aCount = aCounts.reduce(ee.Reducer.sum(), [0]).get([0])
        aMeans = means.slice(0, 0, i)
        aMean = (
            aMeans.multiply(aCounts)
            .reduce(ee.Reducer.sum(), [0])
            .get([0])
            .divide(aCount)
        )
        bCount = total.subtract(aCount)
        bMean = sums.subtract(aCount.multiply(aMean)).divide(bCount)
        return aCount.multiply(aMean.subtract(mean).pow(2)).add(
            bCount.multiply(bMean.subtract(mean).pow(2))
        )

    bss = indices.map(bss_function)
    output = means.sort(bss).get([-1])
    return ee.Number(output)


def kmeans_extent(
    img,
    hand,
    initial_threshold=0,
    region=None,
    band=None,
    n_samples=500,
    seed=7,
    invert=False,
    scale=90,
):
    """Water thresholding methodology using image values and HAND.
    Method from https://doi.org/10.1016/j.rse.2020.111732

    args:
        img (ee.Image): input image to thresholding algorithm
        hand (ee.Image): Height Above Nearest Drainage image used as axis in clustering
        initial_threshold (float, optional): initial estimate of water/no-water for stratified sampling. default = 0
        region (ee.Geometry | None, optional): region to sample values for KMeans clustering, if set to `None` will use img.geometry(). default = None
        band (str | None,optional): band name to use for thresholding, if set to `None` will use first band in image. default = None
        samples (int, optional): number of stratified samples to gather for clustering from the initial water/no-water map. default=500
        seed (int, optional): random number generator seed for sampling. default = 7
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90

    returns:
        ee.Image: clustered image from KMeans clusterer. Classes assumed to be water/no-water

    """

    def _cluster_center(x):
        return ee.Number(
            (
                band_arr.mask(classes.eq(ee.Number(x))).reduce(ee.Reducer.mean(), [0])
            ).get([0])
        )

    if region is None:
        region = img.geometry()

    if band is None:
        img = img.select([0])
        band = ee.String(img.bandNames().get(0))

    else:
        img = img.select(band)

    hand_band = ee.String(hand.bandNames().get(0))

    # convert invert to a number 0 or 1 for ee server-side
    invert = 1 if invert else 0

    img = img.addBands(hand.unmask(0))

    strata = img.select(band).gt(initial_threshold).rename("strata")

    samples = img.addBands(strata).stratifiedSample(
        numPoints=n_samples,
        classBand="strata",
        region=region,
        scale=scale,
        classValues=[0, 1],
        classPoints=[n_samples, n_samples],
        tileScale=16,
        seed=seed,
    )

    clusterer = ee.Clusterer.wekaKMeans(2, 1).train(samples, [band, hand_band])

    samples = samples.cluster(clusterer, "classes")
    classes = samples.aggregate_array("classes")
    unique = classes.distinct().sort()
    classes = ee.Array(classes)
    band_arr = ee.Array(samples.aggregate_array(band))

    class_means = unique.map(_cluster_center)

    min_mean = class_means.reduce(ee.Reducer.min())

    water_class = class_means.indexOf(min_mean)

    water = img.cluster(clusterer)

    water = ee.Image(ee.Algorithms.If(water_class.eq(0), water.Not(), water))

    return water.rename("water").uint8()


def multidim_semisupervised(
    img,
    bands,
    rank_band=None,
    ranking="min",
    region=None,
    n_samples=500,
    seed=7,
    scale=90,
    proba_threshold=None,
):
    """Implementation of the Automatic water detection from
    multidimensional hierarchical clustering algorithm.
    Method details: https://doi.org/10.1016/j.rse.2020.112209
    Note: this is a similar method, not exact from the paper

    args:
        img (ee.Image): input image to thresholding algorithm
        bands (list | ee.List): band names to use for the semi supervised classification
        rank_band (str, optional): band name used to rank which unserpervised class is water. If None then first band name in `bands` is used. default = None
        ranking (str, optional): method to rank the classes by `rank_band`. Options are 'min' or 'max'. If 'min', then the lowest class mean is considered water. default = 'min'
        region (ee.Geometry | None, optional): region to sample values for KMeans clustering, if set to `None` will use img.geometry(). default = None
        samples (int, optional): number of stratified samples to gather for clustering from the initial water/no-water map. default=500
        seed (int, optional): random number generator seed for sampling. default = 7
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90
        proba_threshold (float, optional): probability threshold to create a binary water mask, should be in range of 0-1. If None, then the probability values are returned. default = None
    """

    if region is None:
        region = img.geometry()

    if bands is None:
        bands = img.bandNames()
    else:
        bands = ee.List(bands)

    if rank_band is None:
        rank_band = ee.String(bands.get(0))

    samples = img.select(bands).sample(
        region=region,
        scale=scale,
        numPixels=n_samples,
        seed=seed,
        dropNulls=True,
        tileScale=16,
    )

    classifier = ml.unsupervised_rf(
        100, samples, features=bands, rank_feature=rank_band, ranking=ranking
    )

    probas = img.select(bands).classify(classifier)

    if proba_threshold is None:
        output = probas.rename("water_proba")
    else:
        output = probas.gt(proba_threshold).rename("water").uint8()

    return output


@decorators.keep_attrs
def fuzzy_otsu(
    img,
    band=None,
    region=None,
    scale=90,
    initial_threshold=-15.5,
    grid_size=0.1,
    max_boxes=20,
    seed=0,
    max_buckets=255,
    min_bucket_width=0.001,
    max_raw=1e6,
):
    """Implementation of Otsu thresholding algorithm.
    Segment the grids into water and land representation using initial threshold and randomly select features to calculate
    minimum and maximum threshold of the image. Calculate Midpoint of min and max threshold using Fuzzy_Gaussian to map water.


    args:
        img (ee.Image): input image to thresholding algorithm
        band (str | None,optional): band name to use for thresholding, if set to `None` will use first band in image. default = None
        region (ee.Geometry | None, optional): region to determine threshold, if set to `None` will use img.geometry(). default = None
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90
        initial_threshold (float, optional): initial estimate of water/no-water for estimating the probabilities of classes in segment. default = -15.5
        grid_size (float, optional): size in decimal degrees to tile image/region to check for bimodality. default = 0.1
        max_boxes (int, optional): maximum number of tiles/boxes to use when determining threshold. default = 20
        seed (int, optional): random number generator seed for randomly selected max_boxes. default = 7
        max_buckets (int, optional): The maximum number of buckets to use when building a histogram; will be rounded up to a power of 2. default = 255
        min_bucket_width (float, optional): The minimum histogram bucket width to allow any power of 2. default = 0.001
        max_raw (int, optional): The number of values to accumulate before building the initial histogram. default = 1e6

    returns:
        ee.Image: thresholded water image.
    """

    # Function to Segment Land and Water Grids
    def segment_grid(feature):
        counts = initImg.reduceRegion(
            reducer=ee.Reducer.sum(),
            geometry=feature.geometry(),
            bestEffort=True,
            scale=scale,
        )

        return feature.set(counts)

    if region is None:
        region = img.geometry()

    if band is None:
        img = img.select([0])
        histBand = ee.String(img.bandNames().get(0))

    else:
        histBand = ee.String(band)
        img = img.select(histBand)

    grid = geeutils.tile_region(region, intersect_geom=region, grid_size=grid_size)

    # Prepare good representation of water and land
    initWater = img.lt(initial_threshold).rename("water")
    initImg = initWater.addBands(initWater.Not().rename("land"))

    grid = grid.map(segment_grid)

    # Select water and land grid
    selection = grid.filter(
        ee.Filter.And(ee.Filter.gt("water", 1000), ee.Filter.gt("land", 1000))
    )

    # Randomly select features
    selection = selection.randomColumn("random")
    nBoxes = ee.Number(selection.size())
    randomSelection = ee.Number(max_boxes).divide(nBoxes)
    selection = selection.filter(ee.Filter.lt("random", randomSelection))

    # Create histogram for selected features to calculate min threshold
    histogram = img.reduceRegion(
        ee.Reducer.histogram(max_buckets, min_bucket_width, max_raw)
        .combine("mean", None, True)
        .combine("variance", None, True),
        selection.geometry(),
        scale,
        bestEffort=True,
        tileScale=16,
    )

    min_threshold = otsu(histogram.get(histBand.cat("_histogram")))

    # Create histogram to calculate max threshold
    histogram2 = img.updateMask(img.lt(min_threshold)).reduceRegion(
        ee.Reducer.histogram(max_buckets, min_bucket_width, max_raw)
        .combine("mean", None, True)
        .combine("variance", None, True),
        selection.geometry(),
        scale,
        bestEffort=True,
        tileScale=16,
    )

    max_threshold = otsu(histogram2.get(histBand.cat("_histogram")))

    # Calculate Midpoint of min and max threshold using Fuzzy_Gaussian
    midpoint = ee.Number(ee.Number(min_threshold).add(ee.Number(max_threshold))).divide(
        2
    )
    spread = 0.2

    gauss = fuzzy.fuzzy_gaussian(img, midpoint, spread).clip(region)

    waterImg = img.lt(midpoint)

    waterImg = waterImg.where(waterImg.eq(0), gauss)

    return ee.Image(waterImg)


@decorators.keep_attrs
def modis_flood(img):
    """Implementation of the NASA  MODIS Flood algorithm. Expects that imagery
    has

    https://cdn.earthdata.nasa.gov/conduit/upload/17162/MCDWD_UserGuide_RevB.pdf

    args:
        img (ee.Image): image to apply algorithm to, should be either MODIS or VIIRS sensor data

    returns:
        ee.Image: resulting water map

    """
    # coefficients for the algorithm
    a = 0.0013500
    b = 0.10811
    c = 0.7
    d = 0.2027
    e = 0.06757

    water = img.expression("((nir+a)/(red+b) < c) and (red < d) and (swir2 < e)",{
        "red": img.select("red"),
        "nir": img.select("nir"),
        "swir2": img.select("swir2"),
        "a": a,
        "b": b,
        "c": c,
        "d": d,
        "e": e
    }).uint8().rename("water")

    return water



import ee
import functools


def keep_attrs(func):
    """Decorator function to set the properties of an image from computations to that of the input
    Function to decorate should take an ee.Image object and return an ee.Image object

    args:
        func (object): function object to wrap. Expects that an element within args is of type
            ee.Image and will use first ee.Image to carry metadata

    Example:
        ```python
        @decorators.carry_metadata
        def ndvi(img):
            return img.normalizedDifference([b1,b2])
        ```
        Returned image(s) will have all of the same metadata properties as the input including `system:time_start`
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # expects an element within args is img
        # will set the metadata to first ee.Image instance
        # this assumption is true for 99% of functions used for ee.ImageCollection.map()
        result = ee.Image(func(*args, **kwargs))
        img = [i for i in args if isinstance(i, ee.Image)][0]
        return img.select().addBands(result)

    return wrapper


def keep_names(func):
    """Decorator function to set the band names of an image from computations to that of the input
    Function to decorate should take an ee.Image object and return an ee.Image object

    args:
        func (object): function object to wrap. Expects that an element within args is of type
            ee.Image and will use first ee.Image to carry metadata

    Example:
        ```python
        @decorators.keep_names
        def my_computation(img):
            const = ee.Image.constant(2)
            return const.multiply(img)
        ```
        Returned image(s) will have the same band names as the input.
        This should only be used if the input and output images will have the same number of bands!
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # expects an element within args is img
        # will set the metadata to first ee.Image instance
        # this assumption is true for 99% of functions used for ee.ImageCollection.map()
        result = ee.Image(func(*args, **kwargs))
        img = [i for i in args if isinstance(i, ee.Image)][0]
        return result.rename(img.bandNames())

    return wrapper



import ee
import math
import copy
from pipetools import pipe
from functools import partial
from hydrafloods import decorators, datasets


def add_time_band(img, offset="year", apply_mask=False):
    """Function to add time band to image. Expects image has `system:time_start` property.
    Added band will have name "time"

    args:
        img (ee.Image): image with `system:time_start` to add time band too
        offset (str, optional): units to calculate from 1970-01-01. default = "year"
        apply_mask (bool, optional): boolean switch to apply image mask to time band. if False, then time band will have full coverage

    returns:
        ee.Image: image object with time band added
    """
    t = img.date()
    tDiff = t.difference(ee.Date("1970-01-01"), offset)
    time = ee.Image(tDiff).float().rename("time")
    if apply_mask:
        time = time.updateMask(img.select([0]).mask())
    return img.addBands(time)


def prep_inputs(collection, keep_bands=None, apply_mask=False):
    """Helper function to prepare inputs into time series algorithms.
    Will check if each image in collection has a time and constant band, if not it will add to collection
    Wraps `add_time_band()` for adding the time band

    args:
        collection (ee.ImageCollection): image collection to check and add time/constant band too
        keep_bands (list[str] | None, optional): regex name or list of band names to drop during prep and include in the result
            Will check if "time" or "constant" in list, if not then add to collection. default = None
        apply_mask (bool, optional): boolean switch to apply image mask to time band. if False, then time band will have full coverage

    returns:
        ee.ImageCollection: collection with time and constant bands included
    """
    if not isinstance(collection, ee.ImageCollection):
        collection = collection.collection
    first = ee.Image(collection.first())
    outCollection = copy.deepcopy(collection)

    if keep_bands is None:
        keep_bands = []

    prepfunc = None

    if "time" not in keep_bands:
        prepfunc = partial(add_time_band, apply_mask=apply_mask)
        # outCollection = outCollection.map(tband)
    if "constant" not in keep_bands:
        add_const = lambda x: x.addBands(ee.Image(1))
        if "time" in keep_bands:
            prepfunc = add_const
        else:
            prepfunc = pipe | prepfunc | add_const

    if ("time" not in keep_bands) or ("constant" not in keep_bands):
        outCollection = outCollection.map(prepfunc)

    out_band_order = ee.List(["constant", "time"]).cat(keep_bands)
    return outCollection.select(out_band_order)


def _get_names(base, n):
    """Private helper function to get band names based on base and number of iterations
    i.e. band_1, band_2, ... , band_n
    """
    return ee.List([f"{base}_{i}" for i in range(1, n + 1)])


def add_harmonic_coefs(image, n_cycles=2):
    """Function to add harmonic coefficients as bands to images
    Harmonic coefficients are calculated as sin and cos of frequency within year

    args:
        image (ee.Image): image object to add harmonic coefficiencts to. Expects that image has time band
        n_cycles (int, optional): number of interannual cycles to include. default = 2

    returns:
        ee.Image: image with harmonic coefficient bands added
    """
    cosNames = _get_names("cos", n_cycles)
    sinNames = _get_names("sin", n_cycles)
    frequencyImg = ee.Image.constant(ee.List.sequence(1, n_cycles))
    timeRadians = image.select("time").multiply(2 * math.pi)
    cosines = timeRadians.multiply(frequencyImg).cos().rename(cosNames)
    sines = timeRadians.multiply(frequencyImg).sin().rename(sinNames)

    return image.addBands(cosines).addBands(sines)


def fit_linear_trend(
    collection,
    independents=["constant", "time"],
    dependent=None,
    regression_method="ols",
    output_err=False,
):
    """Function to fit a linear reducer on image collection along the time dimension.

    args:
        collection (ee.ImageCollection): image collection to fit trend line for
        independents (list[str], optional): list of band names to use for fitting regression. default = ["constant", "time"]
        dependent (str | None, optional): band name of values to fit, if None then uses the first band. default = None
        regression_method (str, optional): name of regression reducer to use. options are "simple", "ols", "robust", "ridge", "sen". default = "ols"
        output_err (bool, optional): switch to output regression x and y errors, if true output will have "mean_x", "residual_x" and "residual_y" bands.
            Useful for estimating confidence intervals. default = False

    returns:
        ee.Image: output image with regression coeffients as bands named "offset" and "slope".
            Will have "mean_x", "residual_x" and "residual_y" if ouput_err == True
    """

    independents = ee.List(independents)

    if dependent is None:
        dependent = ee.String(ee.Image(collection.first()).bandNames().get(0))
    else:
        dependent = ee.String(dependent)

    methods = ["simple", "ols", "robust", "ridge", "sen"]
    reducers = [
        ee.Reducer.linearFit(),
        ee.Reducer.linearRegression(independents.length(), 1),
        ee.Reducer.robustLinearRegression(independents.length(), 1),
        ee.Reducer.ridgeRegression(independents.length(), 1),
        ee.Reducer.sensSlope(),
    ]
    method_lookup = {v: reducers[i] for i, v in enumerate(methods)}
    if regression_method not in methods:
        raise ValueError()
    else:
        lin_reducer = method_lookup[regression_method]

    if not isinstance(collection, ee.ImageCollection):
        collection = collection.collection

    reduction_coll = prep_inputs(
        collection, keep_bands=[dependent], apply_mask=True
    ).select(independents.add(dependent))

    n = reduction_coll.select(dependent).reduce("count").rename("n")

    if regression_method in ["sen", "ridge"]:
        independents = ee.List(["constant"]).cat(independents)

    if regression_method == "sen":
        lr = reduction_coll.reduce(lin_reducer, 16).select(
            ["offset", "slope"], independents
        )
    else:
        lr_arr = reduction_coll.reduce(lin_reducer, 16)
        lr = (
            lr_arr.select(["coefficients"])
            .arrayProject([0])
            .arrayFlatten([independents])
        )

    if output_err and regression_method != "sen":
        linear_mse = (
            lr_arr.select("residuals").arrayProject([0]).arrayFlatten([["residual_y"]])
        )

        t_mean = reduction_coll.select("time").mean().rename("mean_x")
        x_err = (
            reduction_coll.select("time")
            .map(lambda x: x.subtract(t_mean).pow(2))
            .reduce(ee.Reducer.sum(), 16)
            .rename("residual_x")
        )

        lr = ee.Image.cat([lr, linear_mse, t_mean, x_err])

    return lr.addBands(n)


def fit_harmonic_trend(
    collection,
    n_cycles=2,
    independents=["constant", "time"],
    dependent=None,
    output_err=False,
):
    """Function to fit a harmonic trend on image collection along the time dimension.
    Uses ee.Reducer.linearRegression to solve for coefficients

    args:
        collection (ee.ImageCollection | hydrafloods.Dataset): image collection to fit trend line for
        n_cycles (int, optional): number of interannual cycles to model. default = 2
        independents (list[str], optional): list of band names to use for fitting regression. default = ["constant", "time"]
        dependent (str | None, optional): band name of values to fit, if None then uses the first band. default = None
        output_err (bool, optional): switch to output regression x and y errors, if true output will have "mean_x", "residual_x" and "residual_y" bands.
            Useful for estimating confidence intervals. default = False

    returns:
        ee.Image: output image with regression coeffients as bands named "constant", "time", "cos_n", and "sin_n" where n is a sequnces of cycles.
            Will have "mean_x", "residual_x" and "residual_y" if ouput_err == True

    raises:
        ValueError: if collection is not of type ee.ImageCollection or hydrafloods.Dataset
    """
    if not isinstance(collection, ee.ImageCollection):
        try:
            collection = collection.collection
        except AttributeError:
            raise ValueError(
                "collection argument expected type ee.ImageCollection or hydrafloods.Dataset,"
                + f"got {type(collection)}"
            )

    if dependent is None:
        dependent = ee.String(ee.Image(collection.first()).bandNames().get(0))
    else:
        dependent = ee.String(dependent)

    independents = (
        ee.List(independents)
        .cat(_get_names("cos", n_cycles))
        .cat(_get_names("sin", n_cycles))
    )

    _add_coefs = partial(add_harmonic_coefs, n_cycles=n_cycles)

    harmonic_collection = prep_inputs(
        collection, keep_bands=[dependent], apply_mask=True
    ).map(_add_coefs)

    harmonic_trend = harmonic_collection.select(independents.add(dependent)).reduce(
        ee.Reducer.linearRegression(numX=independents.length(), numY=1), 16
    )

    n = harmonic_collection.select(dependent).reduce(ee.Reducer.count(), 16).rename("n")

    # Turn the array image into a multi-band image of coefficients
    harmonic_coefficients = (
        harmonic_trend.select("coefficients")
        .arrayProject([0])
        .arrayFlatten([independents])
    ).addBands(n)

    if output_err:
        harmonic_mse = (
            harmonic_trend.select("residuals")
            .arrayProject([0])
            .arrayFlatten([["residual_y"]])
        )

        t_mean = (
            harmonic_collection.select("time")
            .reduce(ee.Reducer.mean(), 16)
            .rename("mean_x")
        )
        x_err = (
            harmonic_collection.select("time")
            .map(lambda x: x.subtract(t_mean).pow(2))
            .reduce(ee.Reducer.sum(), 16)
            .rename("residual_x")
        )

        harmonic_coefficients = ee.Image.cat(
            [harmonic_coefficients, harmonic_mse, t_mean, x_err]
        )

    return harmonic_coefficients


def predict_harmonics(
    collection, harmonics, n_cycles=2, independents=["constant", "time"]
):
    """Helper function to apply harmonic trend prediction

    args:
        collection (ee.ImageCollection): image collection to apply prediction on
        harmonics (ee.Image): harmonic coefficient image with coeffiecient bands
        n_cycles (int, optional): number of interannual cycles to model, note n_cycles must equal
            the number of cycle coefficients in `harmonics`. default = 2
        independents (list[str], optional): list of independent band names to use in model. These are
            other than the harmonic coefficient names. default = ["constant", "time"]

    returns
        ee.ImageCollection: image collection with predicted values
    """

    @decorators.keep_attrs
    def _apply_prediction(image):
        """Closure function to apply prediction"""
        return (
            image.select(independents)
            .multiply(harmonics)
            .reduce("sum")
            .rename("predicted")
        )

    independents = (
        ee.List(independents)
        .cat(_get_names("cos", n_cycles))
        .cat(_get_names("sin", n_cycles))
    )

    _add_coefs = partial(add_harmonic_coefs, n_cycles=n_cycles)

    harmonic_collection = prep_inputs(collection).map(_add_coefs)

    # Compute fitted values.
    predicted = harmonic_collection.map(_apply_prediction)

    return predicted


def get_dummy_img(t):
    """Helper function to get an image readily available to use for predictions
    Resulting image will include time based on year and constant band of 1

    args:
        t (str | ee.Date): string or ee.Date string to create dummy image for

    returns:
        ee.Image: image object with time and constant bands
    """
    if not isinstance(t, ee.Date):
        t = ee.Date(t)
    img = ee.Image.constant(1).set("system:time_start", t.millis())
    time_band = (
        ee.Image(t.difference(ee.Date("1970-01-01"), "year")).float().rename("time")
    )
    return img.addBands(time_band)


def get_dummy_collection(start_time=None, end_time=None, dates=None):
    """Helper function to create image collection of images to apply time series prediction on
    Creates daily imagery between start_time and end_time

    args:
        start_time (str | ee.Date): string or ee.Date string to start creating images on (inclusive)
        end_time (str | ee.Date): string or ee.Date string to end creating images on (exclusive)

    returns:
        ee.ImageCollection: image collection with image containing time and constant bands
    """

    def _gen_image_seq(i):
        t = start_time.advance(ee.Number(i), "day")
        return ee.Image().rename("blank").set("system:time_start", t.millis())

    def _gen_image(d):
        d = ee.Date(d)
        return ee.Image().rename("blank").set("system:time_start", d.millis())

    if (start_time is not None) and (end_time is not None):
        if not isinstance(start_time, ee.Date):
            start_time = ee.Date(start_time)
        if not isinstance(end_time, ee.Date):
            end_time = ee.Date(end_time)

        n = end_time.difference(start_time, "day")
        coll = ee.ImageCollection(ee.List.sequence(0, n).map(_gen_image_seq))
    elif dates is not None:
        coll = ee.ImageCollection(dates.map(_gen_image))

    else:
        raise ValueError("Either 'dates' or 'start_time'/'end_time' need to be defined")

    return prep_inputs(coll)


def temporal_smoothing(collection, reducer, days=10):
    """Function to apply moving window reducer in time on image collection

    args:
        collection (ee.ImageCollection): image collection to apply moving window reducer in time on
        reducer (ee.Reducer): earth engine reducer object to apply
        days (int,optional): size of moving time window in days to apply reducer. default = 10

    returns:
        ee.ImageCollection: image collection with reducer applied in time
    """

    @decorators.keep_attrs
    def _smooth(img):
        """Closure function to apply smoothing in between window"""
        t = img.date()
        band_names = img.bandNames()
        t_start = t.advance(-days // 2, "day")
        t_stop = t.advance(days // 2, "day")
        return (
            collection.filterDate(t_start, t_stop).reduce(reducer, 8).rename(band_names)
        )

    return collection.map(_smooth)


def temporal_iqr_filter(collection):
    """Function to filter values outside of IQR in time on image collection

    args:
        collection (ee.ImageCollection): image collection to apply filter in time on

    returns:
        ee.ImageCollection: image collection with values filtered
    """

    @decorators.keep_attrs
    def _filter(img):
        """Closure function to apply smoothing in between window"""
        mask = img.gt(low_bounds).And(img.lt(upper_bounds))
        return img.updateMask(mask)

    percentiles = collection.reduce(ee.Reducer.percentile([25, 75]))
    iqr = percentiles.select(1).subtract(percentiles.select(0))
    low_bounds = percentiles.select(0).subtract(iqr.multiply(1.5))
    upper_bounds = percentiles.select(1).add(iqr.multiply(1.5))

    return collection.map(_filter)


def anomalies(collection, baseline):
    """Function to calculate anomalies of"""

    @decorators.keep_attrs
    def img_anomaly(img):
        return img.subtract(baseline)

    return collection.map(img_anomaly)



import ee
from collections import OrderedDict
from hydrafloods import decorators


@decorators.keep_attrs
def apply_fcnn(
    image,
    project_name,
    model_name,
    model_kwargs=None,
    output_probas=False,
    output_names=None,
):
    """
    args:
        image (ee.Image): input image for FCNN model, must have all of the features as bands
        project_name (str): cloud project name to reference the model
        model_name (str): ai platform model name
        model_kwargs (dict, optional): dictionary of keyword arguments to pass to ee.Model. default = None
        output_probas (bool, optional): flag to set the output image as class probabilities. If False
            then the ouput will be a one band output of the classification. default = False
        output_bands (Iterable, optional): list of band names to set for the output image
    """

    if model_kwargs is None:
        model_kwargs = OrderedDict(projectName=project_name, modelName=model_name)
    else:
        positional = OrderedDict(projectName=project_name, modelName=model_name)

        model_kwargs = OrderedDict({**positional, **model_kwargs})

    # Load the trained model and use it for prediction.
    model = ee.Model.fromAiPlatformPredictor(**model_kwargs)

    # run the predictions
    predictions = model.predictImage(image.toFloat().toArray())

    if output_probas:
        if output_names is None:
            raise ValueError(
                "please provide `output_names` when `ouput_probas` is set to True"
            )

        output = predictions.arrayFlatten([output_names])

    else:
        output_names = "classification" if output_names is None else output_names
        # find highest probability class
        output = predictions.arrayArgmax().arrayFlatten([[output_names]])

    return output


def minmax_scaling_dict(fc, feature_names):
    """Function to calculate the minimum and maximum values of feautures in a collection
    Expects that fc has all feature names

    args:
        fc (ee.FeatureCollection): feature collection with the features used to calculate min/max value
        feature_names (list[str]):  names of feature columns to calculat min/max values from

    returns
        ee.Dictionary: dictionary of minimum and maximum values for each feature name
    """

    # force ee types
    fc = ee.FeatureCollection(fc)
    ee_feature_names = ee.List(feature_names)

    # apply reducer on each feature column
    feature_min_max = fc.reduceColumns(
        ee.Reducer.minMax().repeat(ee_feature_names.length()), ee_feature_names
    )

    # min/max feature names
    names = ee_feature_names.map(
        lambda x: ee.List([ee.String(x).cat("_min"), ee.String(x).cat("_max")])
    ).flatten()

    # get the min/max values for each feature
    # used to scale values from 0-1
    min_max_dict = ee.Dictionary.fromLists(
        names,
        ee.List(feature_min_max.get("min")).zip(feature_min_max.get("max")).flatten(),
    )

    return min_max_dict


def standard_scaling_dict(fc, feature_names):
    """Function to calculate the mean and standard deviation values of feautures in a collection
    Expects that fc has all feature names

    args:
        fc (ee.FeatureCollection): feature collection with the features used to calculate mean/std dev value
        feature_names (list[str]): names of feature columns to calculat mean/std dev values from

    returns
        ee.Dictionary: dictionary of mean and standard deviation values for each feature name
    """
    # force ee types
    fc = ee.FeatureCollection(fc)
    ee_feature_names = ee.List(feature_names)

    # get a combined reducer for caluclating mean and standard dev
    mean_stddev = ee.Reducer.mean().combine(ee.Reducer.stdDev(), None, True)

    # apply reducer on each feature column
    feature_mean_stddev = fc.reduceColumns(
        mean_stddev.repeat(ee_feature_names.length()), ee_feature_names
    )

    # mean / std dev feature names
    names = ee_feature_names.map(
        lambda x: ee.List([ee.String(x).cat("_mean"), ee.String(x).cat("_stdDev")])
    ).flatten()

    # get the mean / std dev values for each feature
    # used to scale values from ~ -3 to 3
    mean_stddev_dict = ee.Dictionary.fromLists(
        names,
        ee.List(feature_mean_stddev.get("mean"))
        .zip(feature_mean_stddev.get("stdDev"))
        .flatten(),
    )

    return mean_stddev_dict


def minmax_feature_scaling(fc, scaling_dict, feature_names):
    """Function to apply min/max scaling to feature collection

    args:
        fc (ee.FeatureCollection): feature collection to scale
        scaling_dict (ee.Dictionary): dictionary of min/max values to scale to
        feature_names (list[str]):  names of feature columns to calculate apply scaling to

    returns:
        ee.FeatureCollection: scaled feature collection
    """

    def feature_scaling(feature):
        """Nested closure function to apply scaling on each column in each feature"""

        def iter_cols(i):
            """Loops through feature columns"""
            i = ee.String(i)
            v = ee.Number(feature.get(i))
            minv = ee.Number(scaling_dict.get(i.cat("_min")))
            maxv = ee.Number(scaling_dict.get(i.cat("_max")))
            return v.subtract(minv).divide(maxv.subtract(minv))

        # apply scaling on each column of feature
        scaled = ee_feature_names.map(iter_cols)
        # get a dictionary of new values with old feature names
        newVals = ee.Dictionary.fromLists(ee_feature_names, scaled)
        # set feature columns new values
        return feature.set(newVals)

    # force ee types
    fc = ee.FeatureCollection(fc)
    ee_feature_names = ee.List(feature_names)

    # normalize the features in the entire featureCollection
    fc_norm = fc.map(feature_scaling)

    return fc_norm


@decorators.keep_attrs
def minmax_image_scaling(image, scaling_dict, feature_names):
    """Function to scale image between min/max values
    Expects that scaling_dict keys match bands

    args:
        image (ee.Image): image to scale
        scaling_dict (ee.Dictionary): dictionary of min/max values to scale to

    returns
        ee.Image: scaled image
    """
    # get dict as image
    scaling_img = scaling_dict.toImage()
    # extract the min/max values per band
    min_img = scaling_img.select(".*_min")
    max_img = scaling_img.select(".*_max")
    # apply scaling
    return (
        image.select(sorted(feature_names))
        .subtract(min_img)
        .divide(max_img.subtract(min_img))
        .float()
    )


def standard_feature_scaling(fc, scaling_dict, feature_names):
    """Function to apply standard (Z-score) scaling to feature collection

    args:
        fc (ee.FeatureCollection): feature collection to scale
        scaling_dict (ee.Dictionary): dictionary of mean/std dev values for scaling
        feature_names (list[str]):  names of feature columns to calculate apply scaling to

    returns:
        ee.FeatureCollection: scaled feature collection
    """

    def feature_scaling(feature):
        """Nested closure function to apply scaling on each column in each feature"""

        def iter_cols(i):
            """Loops through feature columns"""
            i = ee.String(i)
            v = ee.Number(feature.get(i))
            mean = ee.Number(scaling_dict.get(i.cat("_mean")))
            stddev = ee.Number(scaling_dict.get(i.cat("_stdDev")))
            return v.subtract(mean).divide(stddev)

        # apply scaling on each column of feature
        scaled = ee_feature_names.map(iter_cols)
        # get a dictionary of new values with old feature names
        newVals = ee.Dictionary.fromLists(ee_feature_names, scaled)
        # set feature columns new values
        return feature.set(newVals)

    # force ee types
    fc = ee.FeatureCollection(fc)
    ee_feature_names = ee.List(feature_names)

    # normalize the features in the entire featureCollection
    fc_norm = fc.map(feature_scaling)

    return fc_norm


@decorators.keep_attrs
def standard_image_scaling(image, scaling_dict, feature_names):
    """Function to apply z-score scaling to image
    Expects that scaling_dict keys match bands

    args:
        image (ee.Image): image to scale
        scaling_dict (ee.Dictionary): dictionary of mean/std dev values to scale to

    returns
        ee.Image: scaled image
    """
    # get dict as image
    scaling_img = scaling_dict.toImage()
    # extract the min/max values per band
    mean_img = scaling_img.select(".*_mean")
    stddev_img = scaling_img.select(".*_stdDev")
    # apply scaling
    return (
        image.select(sorted(feature_names))
        .subtract(mean_img)
        .divide(stddev_img)
        .float()
    )


def onehot_feature_encoding(fc, column_name, classes, class_names=None):
    """Function to calculate one-hot encoded columns from categorial columns
    where each new column equals 1 where the class value is the column index

    args:
        fc (ee.FeatureCollection): Feature collection with categorial data to encode
        column_name (str | ee.String): name of column that is categorial to encode
        classes (list[int] | ee.List): list of class values to encode

    kwargs:
        class_names (ee.List, optional): list of names to rename output bands.
            if None then bands will be named b0, b1, ...,bn. default = None

    returns:
        ee.Image: Feature collection with one-hot encoded columns with n new column as n classes
    """

    def feature_encoding(feature):
        c = ee.Number(feature.get(column_name))
        encoded = classes.map(lambda x: c.eq(ee.Number(x)))
        new_cols = ee.Dictionary.fromLists(class_names, encoded)
        return feature.set(new_cols)

    if class_names is None:
        class_names = ee.List.sequence(0, classes.length()).map(
            lambda x: ee.String("b").cat(ee.String(x))
        )

    fc_encoded = fc.map(feature_encoding)

    return fc_encoded


@decorators.keep_attrs
def onehot_image_encoding(img, classes, class_names=None, band=None):
    """Function to convert an categorial image image to one-hot encoded image
    where each new band equals 1 where the class value is the band index

    args:
        img (ee.Image): categorical image to encode
        classes (list[int] | ee.List): list of class values to encode

    kwargs:
        class_names (ee.List, optional): list of names to rename output bands.
            if None then bands will be named b0, b1, ...,bn. default = None
        band (str | ee.String, optional): name of band from input image to endcode.
            if None then the first band is used. default = None

    returns:
        ee.Image: one-hot encoded image with n bands as n classes
    """

    classes = ee.List(classes)

    if class_names is None:
        class_names = ee.List.sequence(0, classes.length()).map(
            lambda x: ee.String("b").cat(ee.String(x))
        )

    if band is None:
        img = img.select([0])
    else:
        img = img.select(band)

    encoded_imgs = classes.map(lambda x: img.eq(ee.Number(x)))

    return (
        ee.ImageCollection.fromImages(ee.List(encoded_imgs))
        .toBands()
        .rename(class_names)
    )


def random_forest_ee(
    n_trees,
    feature_collection,
    feature_names,
    label,
    scaling=None,
    mode="classification",
    min_samples_leaf=1,
):
    """Helper function to scale feature collection and train random forest model

    args:
        n_trees (int): number of trees for random forest model
        feature_collection (ee.FeatureCollection): features to train random forest model
        feature_names (list[str]): names of feature columns to use in random forest model (x values)
        label (str): name of feature column to fit random forest model (y value)
        scaling (str | None, optional): name of scaling to apply before training. One of: "minmax", "standard", `None`.
            default = `None`
        mode (str, optional): The output mode of the random forest model. One of: "classification", "regression",
            "probability". default = "classification"
    """

    if scaling == "minmax":
        scaling_dict = minmax_scaling_dict(feature_collection, feature_names)
        fc_norm = minmax_feature_scaling(
            feature_collection, scaling_dict, feature_names
        )

    elif scaling == "standard":
        scaling_dict = standard_scaling_dict(feature_collection, feature_names)
        fc_norm = standard_feature_scaling(
            feature_collection, scaling_dict, feature_names
        )

    elif scaling is None:
        scaling_dict = None
        fc_norm = feature_collection

    else:
        raise ValueError(
            "Could not determine scaling option. Options are ['minmax', 'standard', or None]"
        )

    classifier = (
        ee.Classifier.smileRandomForest(n_trees, minLeafPopulation=min_samples_leaf)
        .setOutputMode(mode.upper())
        .train(fc_norm, label, feature_names)
    )

    return classifier, scaling_dict


def gradient_boosting_ee(
    n_trees,
    feature_collection,
    feature_names,
    label,
    scaling=None,
    mode="classification",
    shrinkage=0.01,
    loss="LeastAbsoluteDeviation",
):
    """Helper function to scale feature collection and train gradient tree boosting model

    args:
        n_trees (int): number of trees for gradient boosting model
        feature_collection (ee.FeatureCollection): features to train random forest model
        feature_names (list[str]): names of feature columns to use in random forest model (x values)
        label (str): name of feature column to fit random forest model (y value)
        scaling (str | None, optional): name of scaling to apply before training. One of: "minmax", "standard", `None`.
            default = `None`
        mode (str, optional): The output mode of the random forest model. One of: "classification", "regression",
            "probability". default = "classification"
        learning_rate (float,optional): The shrinkage parameter in (0, 1] controls the learning rate of procedure. default = 0.01
        loss (str, optional): Loss function to be optimized. default = "LeastAbsoluteDeviation"
    """

    if scaling == "minmax":
        scaling_dict = minmax_scaling_dict(feature_collection, feature_names)
        fc_norm = minmax_feature_scaling(
            feature_collection, scaling_dict, feature_names
        )

    elif scaling == "standard":
        scaling_dict = standard_scaling_dict(feature_collection, feature_names)
        fc_norm = standard_feature_scaling(
            feature_collection, scaling_dict, feature_names
        )

    elif scaling is None:
        scaling_dict = None
        fc_norm = feature_collection

    else:
        raise ValueError(
            "Could not determine scaling option. Options are ['minmax', 'standard', or None]"
        )

    classifier = (
        ee.Classifier.smileGradientTreeBoost(
            numberOfTrees=n_trees, shrinkage=shrinkage, loss=loss
        )
        .setOutputMode(mode.upper())
        .train(fc_norm, label, feature_names)
    )

    return classifier, scaling_dict


def unsupervised_rf(
    n_trees,
    samples,
    features=None,
    rank_feature=None,
    ranking="min",
):
    """Unserpersived machine learning workflow to classify water
    Methods similar to: https://doi.org/10.1016/j.rse.2020.112209

    args:
        n_trees (int): number of trees to creat random forest model for class generalization
        samples (ee.FeatureCollection): input samples to create water classifier for
        features (list | ee.List): property names from samples to use for the semi supervised classification, If none then all properties are used. default = None
        rank_feature (str, optional): property name used to rank which unserpervised class is water. If None then first band name in `bands` is used. default = None
        ranking (str, optional): method to rank the classes by `rank_band`. Options are 'min' or 'max'. If 'min', then the lowest class mean is considered water. default = 'min'

    returns:
        ee.Classifier.RandomForest: random forest classifier to estimate probability that a pixel is water
    """

    def _cluster_center(x):
        return (
            feature_arr.mask(classes.eq(ee.Number(x))).reduce(ee.Reducer.mean(), [0])
        ).get([0])

    if features is None:
        bands = samples.first().propertyNames()
    else:
        features = ee.List(features)

    if rank_feature is None:
        rank_feature = ee.String(features.get(0))

    clusterer = ee.Clusterer.wekaXMeans(3, 12, 5).train(samples, features)

    samples = samples.cluster(clusterer, "init_classes")

    classes = samples.aggregate_array("init_classes")
    unique = classes.distinct().sort()
    classes = ee.Array(classes)
    feature_arr = ee.Array(samples.aggregate_array(rank_feature))

    class_means = unique.map(_cluster_center)

    if ranking == "min":
        ranker = ee.Reducer.min()
    elif ranking == "max":
        ranker = ee.Reducer.max()
    else:
        raise NotImplementedError(
            "ranking selection is not implemented. options are 'min' or 'max'"
        )

    ranked_mean = class_means.reduce(ranker)

    water_class = class_means.indexOf(ranked_mean)

    binary_samples = samples.map(
        lambda x: (
            ee.Feature(x).set(
                "init_classes", ee.Number(x.get("init_classes")).eq(water_class)
            )
        )
    )

    classifier = (
        ee.Classifier.smileRandomForest(numberOfTrees=n_trees)
        .setOutputMode("PROBABILITY")
        .train(binary_samples, "init_classes", features)
    )

    return classifier


def calc_image_pca(image, region=None, scale=90, max_pixels=1e9, method="svd"):
    """Principal component analysis decomposition of image bands

    args:
        image (ee.Image): image to apply pca to
        region (ee.Geometry | None, optional): region to sample values for covariance matrix,
            if set to `None` will use img.geometry(). default = None
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90
        max_pixels (int, optional): maximum number of pixels to use in reduction operations. default = 1e9
        method (str, optional): the decomposition method for obtaining the eigen vectors and values
            options are 'svd' or 'eigendecomp'. note: svd is usually faster as is does not need to
            compute the covariance matrix of input features. default = 'svd'

    returns:
        ee.Image: principal components scaled by eigen values
    """

    bandNames = image.bandNames()

    out_band_names = ee.List.sequence(1, bandNames.length()).map(
        lambda x: ee.String("pc_").cat(ee.Number(x).int())
    )

    # Mean center the data to enable a faster covariance reducer
    # and an SD stretch of the principal components.
    meanDict = image.reduceRegion(
        reducer=ee.Reducer.mean(), geometry=region, scale=scale, maxPixels=max_pixels
    )
    means = ee.Image.constant(meanDict.values(bandNames))
    centered = image.subtract(means)

    # Collapse the bands of the image into a 1D array per pixel.
    arrays = centered.toArray()

    if method == "svd":
        svd = arrays.toArray(1).matrixSingularValueDecomposition()

        eigen_vecs = svd.select("V")  # .arrayTranspose()
        eigen_vals = svd.select("S").matrixDiagonal()

    elif method == "eigendecomp":
        # Compute the covariance of the bands within the region.
        covar = arrays.reduceRegion(
            reducer=ee.Reducer.centeredCovariance(),
            geometry=region,
            scale=scale,
            maxPixels=max_pixels,
        )

        # Get the 'array' covariance result and cast to an array.
        # This represents the band-to-band covariance within the region.
        covarArray = ee.Array(covar.get("array"))

        # Perform an eigen analysis and slice apart the values and vectors.
        eigens = covarArray.eigen()

        # This is a P-length vector of Eigenvalues.
        eigenValues = eigens.slice(1, 0, 1)
        # This is a PxP matrix with eigenvectors in rows.
        eigenVectors = eigens.slice(1, 1)

    # Convert the array image to 2D arrays for matrix computations.
    arrayImage = arrays.toArray(1)

    # Left multiply the image array by the matrix of eigenvectors.
    principalComponents = ee.Image(eigenVectors).matrixMultiply(arrayImage)

    # Turn the square roots of the Eigenvalues into a P-band image.
    sdImage = (
        ee.Image(eigenValues.sqrt()).arrayProject([0]).arrayFlatten([out_band_names])
    )

    # Turn the PCs into a P-band image, normalized by SD.
    return (
        principalComponents
        # Throw out an an unneeded dimension, [[]] -> [].
        .arrayProject([0])
        # Make the one band array image a multi-band image, [] -> image.
        .arrayFlatten([out_band_names])
        # Normalize the PCs by their SDs.
        .divide(sdImage)
    )


def calc_feature_pca(fc, names, is_centered=False, method="svd"):
    """Principal component decomposition of features

    args:
        fc (ee.FeatureCollection): feature collection to caluculate PCA from
        names (list[str]): property names to uses as features in PCA
        is_centered (bool, optional): boolean to identify if features need to be centered before PCA.
            False means apply centering. default = False
        method (str, optional): the decomposition method for obtaining the eigen vectors and values
            options are 'svd' or 'eigendecomp'. note: svd is usually faster as is does not need to
            compute the covariance matrix of input features. default = 'svd'

    returns:
        ee.Array: eigen vectors of PCA
        ee.Array: eigen values of PCA
        ee.Array: mean values of each feature
    """
    array_ = ee.Array(
        ee.List(
            fc.makeArray(names)
            .aggregate_array("array")
            .map(lambda x: ee.Array(x).toList())
        )
    )
    center = array_.reduce(ee.Reducer.mean(), [0]).repeat(0, array_.length().get([0]))
    if not is_centered:
        centered = array_.subtract(center)
    else:
        centered = array_

    if method == "svd":
        svd = centered.matrixSingularValueDecomposition()

        eigen_vecs = ee.Array(svd.get("V")).transpose()
        eigen_vals = ee.Array(svd.get("S")).matrixDiagonal()

    elif method == "eigendecomp":
        # Compute the covariance of the bands within the region.
        covar = centered.transpose().matrixMultiply(centered)
        # Perform an eigen analysis and slice apart the values and vectors.
        eigens = covar.eigen()

        eigen_vecs = eigens.slice(1, 1)
        eigen_vals = eigens.slice(1, 0, 1)

    else:
        raise ValueError(
            "could not understand provided method keyword. Options are 'svd' or 'eigendecomp'"
        )

    out_band_names = [f"pc_{i}" for i in range(len(names))]

    return eigen_vecs, eigen_vals, center.slice(0, 0, 1).project([1])


def apply_feature_pca(fc, eigen_vecs, names, center=None):
    """Applies Principal component decomposition on features

    args:
        fc (ee.FeatureCollection): feature collection to caluculate pricipal components from
        eigen_vecs (ee.Array): eigen vectors of PCA to transform features
        names (list[str]): property names to uses as features in PCA
        center (ee.Array | None, optional): Array of mean values to center features. If None then no
            centering is applies. default = None

    returns:
        ee.FeatureCollection: feacture collection with new properties within each feature being the principal components
    """
    array_ = ee.Array(
        ee.List(
            fc.makeArray(names)
            .aggregate_array("array")
            .map(lambda x: ee.Array(x).toList())
        )
    )
    if center is not None:
        centered = array_.subtract(
            ee.Array.cat([center], 1).transpose().repeat(0, array_.length().get([0]))
        )
    else:
        centered = array_

    pca_arr = eigen_vecs.matrixMultiply(centered.transpose()).transpose()

    out_band_names = [f"pc_{i}" for i in range(len(names))]

    fc_size = fc.size()
    fc_list = fc.toList(fc_size)
    fc_pca = ee.FeatureCollection(
        ee.List.sequence(0, fc_size.subtract(1)).map(
            lambda x: ee.Feature(fc_list.get(x)).set(
                ee.Dictionary.fromLists(
                    out_band_names,
                    pca_arr.slice(0, x, ee.Number(x).add(1), 1).project([1]).toList(),
                )
            )
        )
    )

    return fc_pca


@decorators.keep_attrs
def apply_image_pca(img, eigen_vecs, names, center=None):
    """Applies Principal component decomposition on image

    args:
        img (ee.Image): image to caluculate pricipal components from
        eigen_vecs (ee.Array): eigen vectors of PCA to transform features
        names (list[str]): band names to uses as features in PCA
        center (ee.Array | None, optional): Array of mean values to center features. If None then no
            centering is applies. default = None

    returns:
        ee.Image: principal components calculated from image
    """
    if center is not None:
        arrayImage = (
            img.select(names)
            .subtract(ee.Image.constant(center.toList()))
            .toArray()
            .toArray(1)
        )
    else:
        arrayImage = img.select(names).toArray().toArray(1)

    principalComponents = ee.Image(eigen_vecs).matrixMultiply(arrayImage)

    out_band_names = [f"pc_{i}" for i in range(len(names))]

    pcaImage = (
        principalComponents
        # Throw out an an unneeded dimension, [[]] -> [].
        .arrayProject([0])
        # Make the one band array image a multi-band image, [] -> image.
        .arrayFlatten([out_band_names])
    )

    return pcaImage


def hist_matching(samples, predictor, target, n_estimators=50):
    """Trains classifiers to perform histogram matching

    args:
        samples (ee.FeatureCollection): feature collection with samples for histogram matching
        predictor (str): column name of values to transform
        target (str): column name of values to match
        n_estimators (int, optional): number of trees to create random forest models from. default = 50

    returns:
        list[ee.Classifier]: list of classifiers with first element being the val to proba and second being proba to val classifiers

    """

    def get_cdf(fc, column):
        def array_to_features(l):
            return ee.Feature(
                None, {column: ee.List(l).get(0), "probability": ee.List(l).get(1)}
            )

        # Histogram equalization start:
        histo = ee.Dictionary(
            fc.reduceColumns(
                ee.Reducer.histogram(
                    maxBuckets=2 ** 12,
                ),
                [column],
            ).get("histogram")
        )

        valsList = ee.List(histo.get("bucketMeans"))
        freqsList = ee.List(histo.get("histogram"))
        cdfArray = ee.Array(freqsList).accum(0)
        total = cdfArray.get([-1])
        normalizedCdf = cdfArray.divide(total)

        array = ee.Array.cat([valsList, normalizedCdf], 1)

        return ee.FeatureCollection(array.toList().map(array_to_features))

    pred_cdf = get_cdf(samples, predictor)
    target_cdf = get_cdf(samples, target)

    proba_to_val = (
        ee.Classifier.smileRandomForest(n_estimators)
        .setOutputMode("REGRESSION")
        .train(
            features=target_cdf, classProperty=target, inputProperties=["probability"]
        )
    )

    val_to_proba = (
        ee.Classifier.smileRandomForest(n_estimators)
        .setOutputMode("REGRESSION")
        .train(
            features=pred_cdf, classProperty="probability", inputProperties=[predictor]
        )
    )

    return val_to_proba, proba_to_val


@decorators.keep_attrs
def apply_image_matching(image, matching_classifiers, output_name="dn"):
    return image.classify(matching_classifiers[0], "probability").classify(
        matching_classifiers[1], output_name
    )



# module to host index functions for image transformations
import ee
from ee.ee_exception import EEException
import math
import string
import random
import datetime
from hydrafloods import decorators


@decorators.keep_attrs
def ndvi(img):
    """Function to calculate Normalized Difference Vegetation Index (NDVI).
    Expects image has "nir" and "red" bands.

    args:
        img (ee.Image): image to calculate NDVI

    returns:
        ee.Image: NDVI image
    """
    return img.normalizedDifference(["nir", "red"]).rename("ndvi")


@decorators.keep_attrs
def evi(img):
    """Function to calculate Enhanced Vegetation Index (EVI).
    Expects image has "blue", "red", and "nir" bands.

    args:
        img (ee.Image): image to calculate EVI

    returns:
        ee.Image: EVI image
    """
    return img.expression(
        "2.5*(nir-red)/(nir+6.0*red-7.5*blue+1)",
        {
            "blue": img.select("blue"),
            "red": img.select("red"),
            "nir": img.select("nir"),
        },
    ).rename("evi")


@decorators.keep_attrs
def mndwi(img):
    """Function to calculate modified Difference Water Index (MNDWI).
    Expects image has "green" and "swir1" bands.

    args:
        img (ee.Image): image to calculate MNDWI

    returns:
        ee.Image: MNDWI image
    """
    return img.normalizedDifference(["green", "swir1"]).rename("mndwi")


@decorators.keep_attrs
def nwi(img):
    """Function to calculate new water index (NWI).
    Expects image has "blue", "nir", "swir1" and "swir2" bands.

    args:
        img (ee.Image): image to calculate NWI

    returns:
        ee.Image: NWI image
    """
    return img.expression(
        "((b-(n+s+w))/(b+(n+s+w)))",
        {
            "b": img.select("blue"),
            "n": img.select("nir"),
            "s": img.select("swir1"),
            "w": img.select("swir2"),
        },
    ).rename("nwi")


@decorators.keep_attrs
def gwi(img):
    """Function to calculate general water index (GWI)
    Expects image has "green", "red", "nir", and "swir1" bands.

    args:
        img (ee.Image): image to calculate GWI

    returns:
        ee.Image: GWI image
    """
    return img.expression(
        "(g+r)-(n+s)",
        {
            "g": img.select("green"),
            "r": img.select("red"),
            "n": img.select("nir"),
            "s": img.select("swir1"),
        },
    ).rename("gwi")


@decorators.keep_attrs
def aewinsh(img):
    """Function to calculate automated water extraction index (AEWI) no shadow
    Expects image has "green", "nir", "swir1" and "swir2" bands.

    args:
        img (ee.Image): image to calculate AEWInsh

    returns:
        ee.Image:  AEWInsh image
    """
    return img.expression(
        "4.0 * (g-s) - ((0.25*n) + (2.75*w))",
        {
            "g": img.select("green"),
            "s": img.select("swir1"),
            "n": img.select("nir"),
            "w": img.select("swir2"),
        },
    ).rename("aewinsh")


@decorators.keep_attrs
def aewish(img):
    """Function to calculate automated water extraction index (AEWI) shadow
    Expects image has "blue", "green", "nir", "swir1" and "swir2" bands.

    args:
        img (ee.Image): image to calculate AEWIsh

    returns:
        ee.Image:  AEWIsh image
    """
    return img.expression(
        "b+2.5*g-1.5*(n+s)-0.25*w",
        {
            "b": img.select("blue"),
            "g": img.select("green"),
            "n": img.select("nir"),
            "s": img.select("swir1"),
            "w": img.select("swir2"),
        },
    ).rename("aewish")


@decorators.keep_attrs
def lswi(img):
    """Function to calculate land surface water index (LSWI).
    Expects image has "nir" and "swir1" bands.

    args:
        img (ee.Image): image to calculate LSWI

    returns:
        ee.Image: LSWI image
    """
    return img.expression(
        "(nir-swir)/(nir+swir)", {"nir": img.select("nir"), "swir": img.select("swir1")}
    ).rename("lswi")


@decorators.keep_attrs
def wri(img):
    """Function to calculate water ratio index (WRI).
    Expects image has "green", "red", "nir" and "swir1" bands.

    args:
        img (ee.Image): image to calculate WRI

    returns:
        ee.Image: WRI image
    """
    return img.expression(
        "(green+red)/(nir+swir)",
        {
            "green": img.select("green"),
            "red": img.select("red"),
            "nir": img.select("nir"),
            "swir": img.select("swir1"),
        },
    ).rename("wri")


@decorators.keep_attrs
def mbwi(img, factor=3):
    """Function to calculate multi band water index (MBWI).
    Expects image has "green", "red", "nir", "swir1", and "swir2" bands.
    https://doi.org/10.1016/j.jag.2018.01.018

    args:
        img (ee.Image): image to calculate MBWI
        factor (int,optional): factor to scale green band for index. default=3

    returns:
        ee.Image: MBWI image
    """
    return img.expression(
        "((factor*green)-red-nir-swir1-swir2)",
        {
            "factor": factor,
            "green": img.select("green"),
            "red": img.select("red"),
            "nir": img.select("nir"),
            "swir1": img.select("swir1"),
            "swir2": img.select("swir2"),
        },
    ).rename("mbwi")


@decorators.keep_attrs
def mwi(img):
    """Function to calculate the Modified Water Index (MWI)
    Expect image has "green", "red", "nir", and "swir1" bands
    https://doi.org/10.1007/978-3-662-45737-5_51 , https://ieeexplore.ieee.org/document/9011209
    """
    # calculate ndvi
    ndvi_img = ndvi(img)
    # calculate mndwi
    mndwi_img = mndwi(img)

    # mwi = 1 - (ndvi - mndwi)
    return ee.Image.constant(1).subtract(ndvi_img.subtract(mndwi_img)).rename("mwi")


@decorators.keep_attrs
def rfi(img):
    """Function to calculate SAR RFI index.
    Expects image has "VV" and "VH" bands.

    args:
        img (ee.Image): image to calculate RFI

    returns:
        ee.Image: RFI image
    """
    return img.expression(
        "(4*VH)/(VV+VH)", {"VV": img.select("VV"), "VH": img.select("VH")}
    ).rename("rfi")


@decorators.keep_attrs
def vv_vh_ratio(img):
    """Function to calculate ratio between VV and VH bands.
    Expects image has "VV" and "VH" bands.

    args:
        img (ee.Image): image to calculate ration

    returns:
        ee.Image: ratio image name 'ratio'
    """
    return img.expression(
        "(VV/VH)", {"VV": img.select("VV"), "VH": img.select("VH")}
    ).rename("ratio")


@decorators.keep_attrs
def vv_vh_abs_sum(img):
    """Function to calculate the absolute value of the sum of VV and VH bands.
    Expects image has "VV" and "VH" bands.

    args:
        img (ee.Image): image to apply calculation

    returns:
        ee.Image: image name 'vv_vh_abs_sum'
    """
    return img.select("VV").add(img.select("VH")).abs().rename("vv_vh_abs_sum")


@decorators.keep_attrs
def ndpi(img):
    """Function to calculate nomalized difference polarization index (NDPI).
    Expects image has "VV" and "VH" bands.

    args:
        img (ee.Image): image to calculate NDPI

    returns:
        ee.Image: NPDI image
    """
    return img.expression(
        "(VV-VH)/(VV+VH)", {"VV": img.select("VV"), "VH": img.select("VH")}
    ).rename("ndpi")


@decorators.keep_attrs
def nvvi(img):
    """Function to calculate nomalized VV index (NVVI).
    Expects image has "VV" and "VH" bands.

    args:
        img (ee.Image): image to calculate NVVI

    returns:
        ee.Image: NVVI image
    """
    return img.expression(
        "(VV)/(VV+VH)", {"VV": img.select("VV"), "VH": img.select("VH")}
    ).rename("nvvi")


@decorators.keep_attrs
def nvhi(img):
    """Function to calculate nomalized VH index (NVHI).
    Expects image has "VV" and "VH" bands.

    args:
        img (ee.Image): image to calculate NVHI

    returns:
        ee.Image: NVHI image
    """
    return img.expression(
        "(VH)/(VV+VH)", {"VV": img.select("VV"), "VH": img.select("VH")}
    ).rename("nvhi")
